#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Naming convention for entities used in for datatoolbox
"""

#%% Emissions
emission_entities = set([
    'Emissions|Kyoto Gases (AR4-GWP100)',
    'Emissions|Kyoto Gases (AR5-GWP100)',
    'Emissions|Kyoto Gases (SAR-GWP100)',
    'Emissions|Kyoto Gases',
    'Emissions|KYOTOGHG SAR',
    'Emissions|KYOTOGHG AR4', # EM|KYO_AR4
    'Emissions|KYOTOGHG AR5', # EM|KYO_AR4
    'Emissions|BC',
    'Emissions|CO2',
    'Emissions|CH4',
    'Emissions|C2F6',
    'Emissions|C6F14',
    'Emissions|CF4',
    'Emissions|NH3',
    'Emissions|N2O',
    'Emissions|NF3',
    'Emissions|NOx',
    'Emissions|HFCs',
    'Emissions|OC',
    'Emissions|SF6',
    'Emissions|PFCs',
    'Emissions|VOC',
    'Emissions|CO',
    'Emissions|Sulfur',
    'Emissions|F-Gases',
    'Emissions|HFC',
    'Emissions|PFC'])

#%% Energy (production if not otherwise stated)
energy_entities = set([
    'Cumulative Capacity', 
    'Capacity Additions',
    'Capacity',
    'Capacity|Electricity|Storage Capacity',
    'Capital Cost',
    'Carbon Sequestration',
    'Primary Energy',           # PE
    'Secondary Energy',         # SE
    'Secondary Energy|Electricity',         # SE
    'Secondary Energy|Heat',         # SE
    'Final Energy',             # FE 
    ]) # ELCT_CAP

#%% Economic enitites
economic_entities = set([
    'Discount rate',
    'GDP|PPP|constant',
    'GDP|PPP|current',
    'GDP|MER',
    'GDP|PPP',
    'Investment',   # INV 
    'Subsidies',    
    'Price',
    'Capital Costs',
    'Policy Cost',
    'Exports',      # EXP
    'Imports',      # IMP
    'Value Added',  # VAL_ADD 
    'Value Lost',   # VA_LO 
    'Population',
    'Trade'])  # POP

#%% Other entities
other_entities = set([
    'Agricultural Production',
    'Area',
    'Count',
    'GMT' # Global Mean Temperature 
    'Climate_Radiative_Forcing',
    'Food Demand',
    'Fertilizer Use',
    'Land Cover',
    'Water Withdrawal',
    'Water Consumption',
    'Forestry Demand',
    'Forestry Production',
    'Yield']) # CRF)

#%% warming assessment
warming_entities = [
    'AR5 climate diagnostics|Concentration',
    'AR5 climate diagnostics|Forcing',
    'AR5 climate diagnostics|Temperature|Exceedance Probability',
    'AR5 climate diagnostics|Temperature',
    'Concentration']
entities = set.union(emission_entities, 
                     energy_entities,
                     economic_entities,
                     other_entities,
                     warming_entities)

# What to do with those? Like pre-fixes?
# Share
# Intensity
# Price
# Concentration
# Production
# Demand
# Storage
# Losses
# Total (implied?)
# Emissions



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Apr 13 22:44:27 2020

@author: ageiges
"""

import git
import pandas as pd
import os
import tqdm 

#%%
path = '/Users/andreasgeiges/Documents/datashelf_minimal'
#path = '/media/sf_Documents/datashelf_minimal'
os.chdir(path)
repo = git.Git(path)

inventory = pd.read_csv(path + "/inventory.csv", index_col=0)

toDelete = list()
for idx,row in tqdm.tqdm(inventory.iterrows()):
#    print(row)
    if  '|' not in idx:
        continue
    
#    parts = idx.split('|')
    source = row['source']
    fileName = row.name + '.csv'
    if os.path.exists('database/' + source + '/' + fileName):
        repo.execute(['git', 'mv', 'database/' + source + '/' + fileName, 'database/' + source + '/' + fileName.replace('|','___')])
    else:
        toDelete.append('database/' + source + '/' + fileName)
        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue May 11 14:36:50 2021

@author: ageiges
"""

import datatoolbox as dt

dt.find(variable ='Final_Energy|Transport|H', source= 'IEA_WEB_2020').index

total = dt.getTable('Final_Energy|Transport|Total__Historic__IEA_WEB_2020')

tables = dict()
for fuel_type in ['Biofuel_waste', 'Electricity']:
    tables[fuel_type] = dt.getTable('Final_Energy|Transport|{}__Historic__IEA_WEB_2020'.format(fuel_type))
    
low_emission_fuel_share = (tables['Biofuel_waste'] + tables['Biofuel_waste']) / total
low_emission_fuel_share  = low_emission_fuel_share.convert('%')
low_emission_fuel_share.meta['source'] = 'CAT calculations based one IEA WEB 2020'
low_emission_fuel_share.loc[['World'],:].to_excel('low_emission_fuel_share_CAT_2021.xlsx')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Apr  3 09:05:54 2020


KAYA identity decomposition
According to :
Sun, J., Ang, B., 2000. Some properties of an exact energy decomposition model. Energy 25, 1177–1188.

Emisisons  =  (emisisons/energy) * (energy/GDP) * (GDP/population) * population
I : Emission intensity per energy
E : Efficiency of energy per GDP
A : Affluence of economy
P : Population
E = I * E * A * P


@author: Andreas Geiges (andreas.geiges@climateanalytics.org)
@reviewer: Gaurav Gandi gaurav.ganti@climateanalytics.org 

"""
#import datatools as dt
import numpy as np
import pandas as pd


colors = ['red', 'green', 'royalblue', 'orange']

#def determineRegion(country):
#    
#    for region in dt.mapping.regions.R5.listAll():
#        
#        if country in dt.mapping.regions.R5.membersOf(region):
#            return region
#
#determineRegion('CHN')
#%%

def _decomposition_to_additive(X, deltaX):
    
    for i in [1,0,2,3]:
        assert X[i].meta['unit'] == deltaX[i].meta['unit'] 
        
    addiditveFactor = \
              deltaX[0] * ( X[1]  * X[2] * X[3]) + \
              deltaX[0] * (1/2) * ( ( deltaX[1] * X[2]      * X[3]) + \
                                    ( X[1]      * deltaX[2] * X[3] ) + \
                                    ( X[1]      * X[2]      * deltaX[3]) )+ \
              deltaX[0] * (1/3) * ( deltaX[1] * deltaX[2] * X[3] + \
                                    deltaX[1] * X[2]      * deltaX[3] + \
                                    X[1]      * deltaX[2] * deltaX[3]) + \
              deltaX[0] * (1/4) * (deltaX[1] *deltaX[2] *deltaX[3]) 
              
    return addiditveFactor

def decomposition_four_factors(X):
    """
    params: 
        X_t0: value of four factors at first timestep t=0
        X_t1: value of four factors at latter timestep t=1
    """
    
    # decomosition works for only for factors
    assert len(X) == 4
#    assert len(X_t1) == 4
    

    
    # computation of differences
    deltaX = [x.yearlyChange() for x in X]
#    for x in X:
#        print(x.meta['unit'] + ' * ', end='')
#    
#    for x in deltaX:
#        print(x.meta['unit'] + ' * ', end='')
#
#    for x in [deltaX[i] for i in [1,0,2,3]]:
#        print(x.meta['unit'] + ' * ', end='')
    

    aX0 = _decomposition_to_additive(X, deltaX)
    aX1 = _decomposition_to_additive([X[i] for i in [1,0,2,3]], [deltaX[i] for i in [1,0,2,3]])
    aX2 = _decomposition_to_additive([X[i] for i in [2,0,1,3]], [deltaX[i] for i in [2,0,1,3]])
    aX3 = _decomposition_to_additive([X[i] for i in [3,0,1,2]], [deltaX[i] for i in [3,0,1,2]])
    
#    aX0.column = [x+1 for x in aX0.columns]
#    aX1.column = [x+1 for x in aX1.columns]
#    aX2.column = [x+1 for x in aX2.columns]
#    aX3.column = [x+1 for x in aX3.columns]
    #print('error: {}'.format(((aX0 + aX1 + aX2 + aX3) - deltaE).loc[:,2016]))
#    import copy
    return ( [aX0, aX1, aX2, aX3])
#    deltaX = [copy.copy(deltaX[i]) for i in [1,0,2,3]]
#    X = [copy.copy(X[i]) for i in [1,0,2,3]]
 
def kaya_decomposion(factors, totalAbs, plot=True, labels = None):
        
        AX = decomposition_four_factors(factors)
    
        if plot:
            kaya_visualization(AX, totalAbs, labels)

def kaya_visualization(AX, totalAbs, labels, relativeChange = False):
    
    import matplotlib.pylab as plt
    
    if len(AX) != len(labels):
        print('inconsitent input')
        
    
    years = AX[0].columns
    countries = AX[0].index
    
    
    
    for country in countries:
        #%%
        plt.figure(num=country)
        plt.clf()
    
        # positive values plotting
        bottom = np.zeros(len(years))
        total  = np.zeros(len(years))
        handles = list()
        for i,factorAbs in enumerate(AX):
        
            if relativeChange:
                factor = 100*(factorAbs / totalAbs)
            else:
                factor = factorAbs
            mask = factor.loc[country,:] >0
            
            h = plt.bar(years[mask], factor.loc[country,mask], bottom=bottom[mask],color=colors[i])
            bottom[mask] += factor.loc[country,mask]
            total[mask]  += factor.loc[country,mask]
            handles.append(h)
        plt.legend(labels)
       
        bottom = np.zeros(len(years))
        for i in range(3,-1,-1):
            if relativeChange:
                factor = 100* (AX[i] /totalAbs)
            else:
                factor = AX[i]
            mask = factor.loc[country,:] <0
            
            plt.bar(years[mask],  factor.loc[country,mask], bottom=bottom[mask],color=colors[i])
            bottom[mask] += factor.loc[country,mask]
            total[mask]  += factor.loc[country,mask]
        plt.plot(years, total, 'kd', markersize=4)
        plt.savefig(country + '_composition.png')
        
    
        
         #%%   
if __name__ == '__main__':
    #%%
    import datatools as dt
    import matplotlib.pylab as plt

    visCountries = ['CHN', 'USA', 'GBR']
    visYears        = list(range(1990,2018))


    populationData = dt.getTable('Population||historic|WDI_2019')#.convert('count')
#    populationData.meta['unit'] = populationData.meta['unit'] + ' / yr'
    gdpData = dt.getTable('GDP|current||historic|WDI_2019').convert('trUSD')
    gdpData.meta['unit'] = gdpData.meta['unit'] + ' / yr'
    energyData = dt.getTable('Final_Energy|Total||historic|IEA_WEB_2019')#.convert('MJ / yr')
    emissionData = dt.getTable('Emissions|Fuel|CO2|Total||historic|IEA_CO2_FUEL_2019')
    emissionData.meta['unit'] = emissionData.meta['unit'] + ' / yr'
    #emissionData = emissionData.convert('kg CO2 / yr')
    #%% computation of the four factors
    intensity = emissionData / energyData
    intensity.meta['entity'] = "emission intensity"
    
    efficiency = (energyData / gdpData)
    efficiency.meta['entity'] = "energy efficiency"
    
    affluence = (gdpData / populationData)
    affluence.meta['entity'] = "affluence (wealth)"
    
    #%% viusualization
    fig = plt.figure('comparison', figsize= [10,10])
    plt.clf()
    fig, ax = plt.subplots(2,2, num='comparison', figsize= [10,10])
    
    intensity.loc[visCountries,visYears].vis.plot(ax= ax[0,0])
    efficiency.loc[visCountries,visYears].vis.plot(ax= ax[0,1])
    affluence.loc[visCountries,visYears].vis.plot(ax= ax[1,0])
    populationData.loc[visCountries,visYears].vis.plot(ax= ax[1,1])
    plt.tight_layout()
    plt.savefig('factor_comparison.png')
    
    #%% test of emission computation
    deltaE = emissionData.yearlyChange().loc[visCountries,visYears]
    emissionsComputed = populationData * affluence * efficiency * intensity
    print('error: {}'.format((emissionsComputed - emissionData).max().max()))

    X = [populationData.loc[visCountries,visYears], affluence.loc[visCountries,visYears], efficiency.loc[visCountries,visYears], intensity.loc[visCountries,visYears]]
    
    AX = decomposition_four_factors(X)
    
    labels = ['Population','GDP/capita','Energy Effiency', 'Emission Intensity']
    
    kaya_visualization(AX, emissionData.loc[visCountries,visYears], labels, relativeChange=True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Aug 19 17:54:38 2020

@author: ageiges
"""

#%%
import datatoolbox as dt
source = 'UNFCCC_CRF_2020'
entities = dt.find(source=source).entity.unique()

for entity in entities:
    dt.find(entity=entity,source=source).graph(entity + '.png')

#%%
res = dt.find(source=source)
res.index
dt.removeTables(list(res.index))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Nov 16 10:43:55 2021

@author: ageiges
"""
import numpy as np
import pandas as pd
import xarray as xr
import pyam
from tqdm import tqdm
from . import config
from time import time
import datatoolbox as dt
# %% private functions
def _get_dimension_indices(table, dimensions):
    
    ind = list()
    for dim in dimensions:
        if isinstance(dim, tuple):
        
            index = [tuple(_get_unique_labels(table, sub_dim)[0] for sub_dim in dim)] # todo find better way
        elif dim == table.index.name:
            index = list(table.index)
        elif dim == table.columns.name:
            index = list(table.columns)
        elif dim in table.attrs.keys():
            index = [table.attrs[dim]]
        ind.append(index)
        
    return ind

def _yearsColumnsOnly(index):
    """
    Extracts from any given index only the index list that can resemble 
    as year 
    
    e.g. 2001
    """
    
    import re
    REG_YEAR   = re.compile('^[0-9]{4}$')
    
    newColumns = []
    for col in index:
        if REG_YEAR.search(str(col)) is not None:
            newColumns.append(col)
        else:
            try: 
                if ~np.isnan(col) and REG_YEAR.search(str(int(col))) is not None:
                #   test float string
                    newColumns.append(col)
            except:
                pass
    return newColumns

def _get_meta_collection(table_iterable, dimensions):
    """

    Parameters
    ----------
    table_iterable : list of tables
        DESCRIPTION.
    dimensions : list of dimentions
        DESCRIPTION.

    Returns
    -------
    metaCollection : TYPE
        DESCRIPTION.

    """

    
    metaCollection = dict()
    for table in table_iterable:
        
        for key in table.meta.keys():
            if key in dimensions  or key == 'ID':
                continue
            if key not in metaCollection.keys():
                metaCollection[key] = set()
                
            metaCollection[key].add(table.meta[key])
    
    return metaCollection

def _get_unique_labels(table, dim):
    
    if isinstance(dim, tuple):
        unique_lables = [tuple(_get_unique_labels(table, sub_dim)[0] for sub_dim in dim)]
        # unique_lables = [tuple(d for sub_dim in dim for d in _get_unique_labels(table, sub_dim))] # todo find better way
    elif dim == table.index.name:
        unique_lables = table.index
    elif dim == table.columns.name:
        unique_lables = table.columns
    elif dim in table.meta.keys():
        unique_lables = [table.meta[dim]]
    else:
        #raise(Exception(f'Dimension {dim} not available'))
        unique_lables = [np.nan]
        
    return unique_lables

def _get_dimensions(table_iterable, dimensions):
    
    
    dims = dict()
    
    for table in table_iterable:
    
        for dim in dimensions:
            
            dims[dim] = dims.get(dim,set()).union(_get_unique_labels(table, dim))
    return dims

def _to_xarray(tables, dimensions, stacked_dims):
    """ 
    Return a database query result as an xarray . This constuctor allows only for
    one unit, since the full array is quantified using pint-xarray. 
    The xarray dimensions (coordiantes) are defined
    by the provided dimensions. A multi-index for a coordinate can be created
    by using stacked_dims.
    
    Usage:
    -------
    tables : Iterable[[dt.Datatable]]
    dimensions :  Iterable[str]]
        Dimensions of the shared yarray dimensions / coordinates
    stacked_dims : Dict[str]]
        Dictionary of all mutli-index coordinates and their sub-dimensions
    
    Returns
    -------
    matches : xarray.Dataset + pint quantification
    """   
    tt = time()
    metaCollection = _get_meta_collection(tables, dimensions)
    if config.DEBUG:
        print(f'ime required for meta collection: {time()-tt:2.2f}s')
    
    
    tt = time()
    final_dims = dimensions.copy()
    xdims = dimensions.copy()
    for st_dim, sub_dims in stacked_dims.items():
        [xdims.remove(dim) for dim in sub_dims]
        xdims.append(sub_dims)
        
        [final_dims.remove(dim) for dim in sub_dims]
        final_dims.append(st_dim)
        
    dims = _get_dimensions(tables, xdims)
    if config.DEBUG:
        print(dims)
    coords = {x: sorted(list(dims[x])) for x in dims.keys()}
    labels = dict()
    for st_dim, sub_dims in stacked_dims.items():
        coords[st_dim] = range(len(coords[sub_dims]))
        sub_labels = list()
        for i_dim, sub_dim in enumerate(sub_dims):
            
            sub_labels.append(pd.Index([x[i_dim] for x in coords[sub_dims]], name=sub_dim))
        labels[st_dim]  = pd.MultiIndex.from_arrays(sub_labels, names = sub_dims)
        del coords[sub_dims]
    
    dimSize = [len(labels) for dim, labels in dims.items()]
    
    if config.DEBUG:
        print(f'Get timension: {time()-tt:2.2f}s')
    
    tt = time()
    xData =  xr.DataArray(np.zeros(dimSize)*np.nan, coords=coords, dims=final_dims).assign_coords(labels)
    
    tt= time()
    for table in tables:
        ind = _get_dimension_indices(table,xdims)
        # xData.loc[tuple(ind)] = table.values.reshape(len(table.index),len(table.columns),1)
        
        xData.loc[tuple(ind)] = table.values.reshape(*[len(x) for x in ind])
    if config.DEBUG:
        print(f'Time required for xr data filling: {time()-tt:2.2f}s')    
    tt=time()
    metaCollection['unit'] = list(metaCollection['unit'])[0]
    xData = xData.pint.quantify(metaCollection['unit']).assign_coords(labels)
    xData.attrs = metaCollection
    
    return xData

def _key_set_to_xdataset(dict_of_data,
                        dimensions = ['model', 'scenario', 'region', 'time'],
                        stacked_dims = {'pathway': ('model', 'scenario')}):
    """ 
    Returns xarry dataset converted from a dict of pandas dataframes  or dt.TableSet. Differenty variables
    are stored as key variables. The xarray dimensions (coordiantes) are defined
    by the provided dimensions. A multi-index for a coordinate can be created
    by using stacked_dims.
    
    Usage:
    -------
    dimensions :  Iterable[str]]
        Dimensions of the shared yarray dimensions / coordinates
    stacked_dims : Dict[str]]
        Dictionary of all mutli-index coordinates and their sub-dimensions
    
    Returns
    -------
    matches : xarray.Dataset + pint quantification
    """    
    
    dim_to_sort = 'variable'
    sort_dict = dict()
    for key, table in dict_of_data.items():
        
        var = table.meta['variable']
        
        if var in sort_dict.keys():
            sort_dict[var].append(table)
        else:
            sort_dict[var] = [table]
            
    
    variables = sort_dict.keys()

    data = list()
    for variable in variables:
        tables = sort_dict[variable]
        
        xarray = _to_xarray(tables, dimensions, stacked_dims)
        data.append(xr.Dataset({variable : xarray}))
        
    ds = xr.merge(data) 

    
    return ds

def _pack_dimensions(index, **stacked_dims):
    packed_labels = {}
    packed_values = {}
    drop_levels = []
    
    for dim, levels in stacked_dims.items():
        labels = pd.MultiIndex.from_arrays([index.get_level_values(l) for l in levels])
        packed_labels[dim] = labels_u = labels.unique()
        packed_values[dim] = pd.Index(labels_u.get_indexer(labels), name=dim)
        drop_levels.extend(levels)

    return (
        pd.MultiIndex.from_arrays(
            [index.get_level_values(l) for l in index.names.difference(drop_levels)] +
            list(packed_values.values())
        ),
        packed_labels
    )


def to_dataset(data_dict,
               var_data_labels, # with different units
               packed_dimensions, # dependent dimenions
               ):
    
    
    
    pass

#%%
def _xDataSet_to_wide_dataframe(xds):
    merge_list = list()
    dims = list(xds.coords)
    assert (dt.config.DATATABLE_COLUMN_NAME in dims) and (dt.config.DATATABLE_INDEX_NAME in dims)
    df = xds.to_dataframe()
    
    for var in xds.var():
        #print(var)
        wdf = df[var]
        level = list(df.index.names).index('time')
        wdf = wdf.unstack(level=level)
        wdf['variable'] = var
        wdf['unit']  = xds[var].attrs['unit']
        wdf.set_index('variable', append=True, inplace=True)
        wdf.set_index('unit', append=True, inplace=True)
        merge_list.append(wdf)
    return pd.concat(merge_list)

def _xDataArray_to_wide_df(xarr):
    df = xarr.to_dataframe()
    level = list(df.index.names).index('year')
    wdf = df.unstack(level=level)
    wdf['variable'] = xarr.name
    wdf['unit']  = str(xarr.pint.units)
    wdf.set_index('variable', append=True, inplace=True)
    wdf.set_index('unit', append=True, inplace=True)
    wdf.columns = wdf.columns.droplevel(0)
    return wdf

def to_pyam(data):
    """
    Converts known data tyes to a tableset. Recognized data types are:
        - xarray.Dataset

    Parameters
    ----------
    data : TYPE
        DESCRIPTION.

    Returns
    -------
    pyam.IamDataFrame

    """
    if isinstance(data ,xr.Dataset):
        wdf = _xDataSet_to_wide_dataframe(data)
        return pyam.IamDataFrame(wdf)
    
    if isinstance(data, xr.DataArray):
        wdf = _xDataArray_to_wide_df(data)
        return pyam.IamDataFrame(wdf)
    
    else:
        raise(Exception(f'{type(data)} is not implemented'))
        
def to_wide_dataframe(data, index_cols = ['Variable', 'Model', 'Scenario']):
    if isinstance(data, pd.DataFrame):
        wdf = data.set_index(index_cols)
        
    elif isinstance(data ,xr.Dataset):
        wdf = _xDataSet_to_wide_dataframe(data)
        # wdf = wdf.reset_index(level='region')
    
    return wdf


def to_tableset(data, additional_meta = dict()):
    """
    Converts known data tyes to a tableset. Recognized data types are:
        - xarray.Dataset

    Parameters
    ----------
    data : TYPE
        DESCRIPTION.

    Returns
    -------
    None.

    """

    
    if isinstance(data, pd.DataFrame):
        years = _yearsColumnsOnly(data)
        meta_cols = data.columns.difference(years)
        data = data.set_index(list(meta_cols))
        data = data.reset_index(level='region')
        ts = dt.TableSet()
        for i,idx in enumerate(data.index.unique()):
            sel = data.loc[idx,:]
            if isinstance(sel, pd.Series):
                sel = pd.DataFrame(sel).T.set_index('region')
            else:
                sel = data.loc[idx,:].set_index('region')
            meta = {x:y for x,y in zip(data.index.names, idx)}
            meta.update(additional_meta)
            try :
                dt.core.getUnit(meta['unit'])
            except:
                print(f'Skipping because of unit {meta["unit"]}') 
                print(meta)
                continue
            table = dt.Datatable(data = sel,
                                 meta = meta)
            ts[i] = table
            
        return ts
    
    
    elif isinstance(data ,xr.Dataset):
        wdf = _xDataSet_to_wide_dataframe(data)
        wdf = wdf.reset_index(level='region')
        ts = dt.TableSet()
        for i,idx in enumerate(wdf.index.unique()):
            sel = wdf.loc[idx,:].set_index('region')
            meta = {x:y for x,y in zip(wdf.index.names, idx)}
            meta.update(additional_meta)
            table = dt.Datatable(data = sel,
                                 meta = meta)
            ts[i] = table
        return ts
    
    elif isinstance(data, pyam.IamDataFrame):
        
        wdf = idf.timeseries().reset_index()
        idx_cols = ['variable','model', 'scenario', 'unit']
        wdf = wdf.set_index(idx_cols)
        tables = dt.Tableset()
        for idx, df in tqdm(wdf.groupby(idx_cols)):
            meta = {key: value for key, value in zip(idx_cols, idx)}
            meta = dt.core._split_variable(meta)
            meta.update(additional_meta)
            try:
                dt.core.ur(meta['unit'])
            except:
                print(f"Skiping table due to unit {meta['unit']}")
                continue
            table = dt.Datatable(df.set_index('region'), meta = meta).clean()
            tables.add(table)
    
    
    else:
        raise(Exception(f'{type(data)} is not implemented'))
       
     
       
def to_xdataset(data, 
                dimensions = ['model', 'scenario', 'region', 'time'],
                stacked_dims = {'pathway': ('model', 'scenario')}):
        
    
     if isinstance(data, pd.DataFrame):
         # wide dataframe
         for dim in dimensions:
             if dim not in data.columns:
                 print(f'Dimension {dim} not found in index names')
             
         #convert to table set
         print('convert to table set')
         data = to_tableset(data)
         
         ds = _key_set_to_xdataset(data)
    
     if isinstance(data, pyam.IamDataFrame):
        
         ds = dt.data_structures.DataSet(data)
    
     return ds
        
 
    #%%
    
    #%%
if __name__ == '__main__':
    
    #%test
    import datatoolbox as dt

    # local_data_file = '/media/sf_Documents/python/ca_data_management/data/IPCC_AR6/AR6_Scenarios_Database_World_v1.0.csv'#,
    # data = pd.read_csv(local_data_file)
    # dimensions = ['Model', 'Scenario', 'Region', 'year']
    # stacked_dims = {'Pathway': ('Model', 'Scenario'),
    #                 'varunit' : ("Variable", "Unit")}
    # data = data.set_index(data.columns[:5].tolist()).rename_axis(columns= 'year').stack()
    
    # from datatoolbox.tools.pyam import idf_to_xarray
    
    # xds = idf_to_xarray(data, stacked_dims= stacked_dims)
    
    # tbs = dt.getTables(dt.find(source='IAMC15_2019_R2').index[:10])
    # idf = tbs.to_IamDataFrame()
    # # xda = idf_to_xarray(idf)
    # stacked_dims= {'pathway': ('model', 'scenario')}
    
    # index, labels = _pack_dimensions(idf.index, **stacked_dims)
    xdata = dt.findp(variable=['Emissions|CO2|Energy|Supply|Electricity',
                              'Secondary Energy|Electricity'],
                    source ='IPCC_SR15', pathway='**SSP1-19**').as_xarray()
    
    ts = to_tableset(xdata)
    idf = to_pyam(xdata)
    
    wdf = to_wide_dataframe(xdata)
    idf2 = dt.findp(variable=['Emissions|CO2|Energy|Supply|Electricity',
                              'Secondary Energy|Electricity'],
                    source ='IPCC_SR15', pathway='**SSP1-19**').as_pyam()
    
    ds = to_xdataset(wdf.reset_index())

#%% 
# res = 


import datatoolbox as dt
import pandas as pd
import numpy as np
import numpy.testing as npt

#%%

metaDict = {'entity':'Emissions|Candy',
                           'category': '',
                           'scenario':'Historic',
                           'source': 'SANDBOX',
                           'unit' : 'Mt CO2'}

metaDict2 = {'entity':'Emissions|Candy',
                           'category': '',
                           'scenario':'Historic',
                           'source': 'SANDBOX',
                           'unit' : 'Gg CO2'}

df1 = dt.Datatable([[10,20,30,],
                   [40,40,50,]], 
                   columns = [2000, 2010, 2020],
                   index= ['DEU', 'IND'],
                   meta = metaDict)

df2 = dt.Datatable([[50.,50,50,],
                   [40,40,40,]], 
                   columns = [2000, 2010, 2020],
                   index= ['DEU', 'IND'],
                   meta = metaDict)

df3 = dt.Datatable([[50,50],
                   [40,40]], 
                   columns = [2000, 2010],
                   index= ['DEU', 'IND'],
                    meta = metaDict2)

def test_addition_int():
    
    # adding integer
    exp = dt.Datatable([[30,40,50,],
                       [60,60,70,]], 
                       columns = [2000, 2010, 2020],
                       index= ['DEU', 'IND'],
                       meta = metaDict)
    
    obs = df1 + 20    
    # TODO use with new pandas version
    # assert_frame_equal(obs, exp, check_dtype=False)
    npt.assert_array_almost_equal(obs, exp, decimal = 6)
    assert obs.meta['unit']   == exp.meta['unit']
    assert obs.meta['source'] == 'calculation'
    
    obs = 20 + df1
    # TODO use with new pandas version
    # assert_frame_equal(obs, exp, check_dtype=False)
    npt.assert_array_almost_equal(obs, exp, decimal = 6)
    assert obs.meta['unit']   == exp.meta['unit']
    assert obs.meta['source'] == 'calculation'

def test_addition_dfs():
        
    # adding two datatables
    obs = df1 + df2
    exp = dt.Datatable([[60,70,80,],
                       [80,80,90,]], 
                       columns = [2000, 2010, 2020],
                       index= ['DEU', 'IND'],
                       meta = metaDict)
    # TODO use with new pandas version
    # assert_frame_equal(obs, exp, check_dtype=False)
    npt.assert_array_almost_equal(obs, exp, decimal = 6)
    assert obs.meta['unit']   == exp.meta['unit']
    assert obs.meta['source'] == 'calculation'

def test_raddition_dfs_conv():
    # adding two datatables + conversion
    obs = df1 + df3
    exp = dt.Datatable([[10.05,  20.05, np.nan],
                       [40.04,  40.04,  np.nan]], 
                       columns = [2000, 2010, 2020],
                       index= ['DEU', 'IND'],
                       meta = metaDict)
    
    # TODO use with new pandas version
    # assert_frame_equal(obs, exp, check_dtype=False)
    npt.assert_array_almost_equal(obs, exp, decimal = 6)
    assert obs.meta['unit']   == exp.meta['unit']
    assert obs.meta['source'] == 'calculation'

    
def test_laddition_dfs_conv():
    
    obs = df3 + df1
    exp = dt.Datatable([[10050.,  20050, np.nan],
                       [40040,  40040,  np.nan]], 
                       columns = [2000, 2010, 2020],
                       index= ['DEU', 'IND'],
                       meta = metaDict2)

    # TODO: conversion makes this a float, fix?
    # TODO use with new pandas version
    # assert_frame_equal(obs, exp, check_dtype=False)
    npt.assert_array_almost_equal(obs, exp, decimal = 6)
    assert obs.loc[:,2020].isnull().all()
    assert obs.meta['unit']   == exp.meta['unit']
    assert obs.meta['source'] == 'calculation'

def test_sum():
    obs = sum([df3, df1])
    exp = dt.Datatable([[10050.,  20050, np.nan],
                       [40040,  40040,  np.nan]], 
                       columns = [2000, 2010, 2020],
                       index= ['DEU', 'IND'],
                       meta = metaDict2)

    # TODO: conversion makes this a float, fix?
    # TODO:use with new pandas version
    #assert_frame_equal(obs, exp, check_dtype=False)
 
    npt.assert_array_almost_equal(obs, exp, decimal = 5)
    assert obs.loc[:,2020].isnull().all()
    assert obs.meta['unit']   == exp.meta['unit']
    assert obs.meta['source'] == 'calculation'  
    
def test_rsubstraction_int():
    
    # adding integer
    exp = dt.Datatable([[-10, 0,10,],
                       [20,20,30,]], 
                       columns = [2000, 2010, 2020],
                       index= ['DEU', 'IND'],
                       meta = metaDict)
    
    obs = df1 - 20
    #assert_frame_equal(obs, exp)
    npt.assert_array_almost_equal(obs, exp, decimal = 5)
    assert obs.meta['unit']   == exp.meta['unit']
    assert obs.meta['source'] == 'calculation'
    
def test_lsubstraction_int():
    
    exp = dt.Datatable([[10, 0,-10,],
                       [-20,-20,-30,]], 
                       columns = [2000, 2010, 2020],
                       index= ['DEU', 'IND'],
                       meta = metaDict)
    
    obs = 20 - df1
    #assert_frame_equal(obs, exp)
    npt.assert_array_almost_equal(obs, exp, decimal = 5)
    assert obs.meta['unit']   == exp.meta['unit']
    assert obs.meta['source'] == 'calculation'

def test_substraction_dfs():
    
    # adding two datatables
    obs = df1 - df2
    exp = dt.Datatable([[-40,-30,-20,],
                       [0,0,10,]], 
                       columns = [2000, 2010, 2020],
                       index= ['DEU', 'IND'],
                       meta = metaDict)
    
    #assert_frame_equal(obs, exp)
    npt.assert_array_almost_equal(obs, exp, decimal = 5)
    assert obs.meta['unit']   == exp.meta['unit']
    assert obs.meta['source'] == 'calculation'

def test_rsubstraction_dfs_conv():

    # adding two datatables + conversion
    
    obs = df1 - df3
    exp = dt.Datatable([[9.95,  19.95, np.nan],
                       [39.96,  39.96,  np.nan]], 
                       columns = [2000, 2010, 2020],
                       index= ['DEU', 'IND'],
                       meta = metaDict)
    

    # TODO:use with new pandas version
    #assert_frame_equal(obs, exp)

    npt.assert_array_almost_equal(obs, exp, decimal = 5)
    assert obs.loc[:,2020].isnull().all()
    assert obs.meta['unit']   == exp.meta['unit']
    assert obs.meta['source'] == 'calculation'

def test_lsubstraction_dfs_conv():
    
    obs = df3 - df1
    exp = dt.Datatable([[-9950.0, -19950.0, np.nan],
                       [-39960.0, -39960.0,  np.nan]], 
                       columns = [2000, 2010, 2020],
                       index= ['DEU', 'IND'],
                       meta = metaDict2)

    # TODO:use with new pandas version
    # assert_frame_equal(obs, exp)
    
    npt.assert_array_almost_equal(obs, exp, decimal = 5)
    assert obs.loc[:,2020].isnull().all()
    assert obs.meta['unit']   == exp.meta['unit']
    assert obs.meta['source'] == 'calculation'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Mar 29 17:25:52 2019

@author: Andreas Geiges
"""
import datatoolbox as dt
import pandas as pd


def initializeSpatialMapping():
    #SETUP
    MAPPING_FILE_PATH = 'data/region_mappings.csv'
    CONTINENT_FILE_PATH = 'data/all.csv'
    
    print('Using file ' + MAPPING_FILE_PATH + ' to create initial mapping')
            
    mappingDf = pd.read_csv(MAPPING_FILE_PATH, index_col=0)
    continentDf = pd.read_csv(CONTINENT_FILE_PATH, index_col=0)
    continentDf = continentDf.set_index('alpha-3')
    
#    for co in mappingDf.index:
#        if co not in dt.geoMapping.countries:
#            print(co)
    mappingDf = mappingDf.loc[dt.mapp.countries.listAll()]
    # EU28
    dt.mapp.addNewRegion('EU28', mappingDf.eu28=='EU')
    
    for continent in continentDf.region.unique():
        if pd.isna(continent):
            continue
        
        dt.geoMapping.addNewRegion(continent, continentDf.region==continent)
    
    dt.geoMapping.commitMapping()
        
def initSourceTable():
    sourceDf = pd.DataFrame(columns = dt.config.SOURCE_META_FIELDS)
    
    sourceDf.to_csv(dt.config.SOURCE_FILE)
    dt.core.DB._gitAddFile(dt.config.SOURCE_FILE)
    dt.core.DB._gitCommit('added empty source file')
    

def initCATRegions():
    rawInput = {'R5OECD' : 'AIAALAANDASMATAATFAUSAUTBELBVTCANCCKCHECHICXRCYMDEUDNKESPFINFJIFRAFROGBRGGYGIBGRCGRLGUMIMNIOTIRLISLITAJEYJPNLIELUXMCONCLNFKNLDNORNZLPRTPYFSJMSLBSMRSPMSWETURUMIUSAVATVGBVIRVUTWSM',
                'R5MAF'  : 'AGOAREBDIBENBFABHRBWACAFCIVCMRCODCOGCOMCPVDJIDZAEGYERIESHETHGABGHAGINGMBGNBGNQIRNIRQISRJORKENKWTLBNLBRLBYLSOMARMDGMLIMOZMRTMUSMWIMYTNAMNERNGAOMNPSEQATREURWASAUSDNSENSHNSLESOMSSDSTPSWZSYRTCDTGOTUNTZAUGAYEMZAFZMBZWE',
                'R5REF'  : 'ALBARMAZEBGRBIHBLRCYPCZEESTGEOHRVHUNKAZKGZLTULVAMDAMKDMLTMNEPOLROURUSSRBSVKSVNTJKTKMUKRUZB',
                'R5ASIA' : 'AFGBGDBRNBTNCHNCOKFSMHKGHMDIDNINDKHMKIRKORLAOLKAMACMDVMHLMMRMNGMNPMYSNIUNPLNRUPAKPHLPLWPNGPRKSGPSYCTHATKLTLSTONTUVTWNVNMWLF',
                'R5LAM'  : 'ABWANTARGATGBESBHSBLMBLZBMUBOLBRABRBCHLCOLCRICUBCUWDMADOMECUFLKGLPGRDGTMGUFGUYHNDHTIJAMKNALCAMAFMEXMSRMTQNICPANPCNPERPRIPRYSGSSLVSURSXMTCATTOURYVCTVEN'}
    mappingToCountries = dict()
    for region in rawInput.keys():
        countryList = [rawInput[region][i:i+3] for i in range(0,len(rawInput[region]),3)]
        mappingToCountries[region] =  countryList
    dt.mapp.regions.createNewContext('CAT',mappingToCountries)
    dt.mapp.regions.save()
    
G20List= [dt.mapp.getSpatialID(x) for x in [ 'Argentina',
                          'Australia',
                          'Brazil',
                          'Canada',
                          'China',
                          'India',
                          'Indonesia',
                          'Japan',
                          'Mexico',
                          'Russian Federation',
                          'Saudi Arabia',
                          'South Africa',
                          'Korea, Republic of',
                          'Turkey',
                          'United Kingdom',
                          'United States']] + list(dt.mapp.regions.EU28.membersOf('EU28'))
politicallDict = {'G20': G20List}

CAT_countries = ['ARE', 'ARG', 'AUS', 'BRA', 'BTN', 'CAN', 'CHE', 'CHL', 'CHN', 'CRI', 'ETH', 'EU28', 'GMB', 'IDN', 'IND', 'JPN', 'KAZ', 'KOR', 'MEX', 'MAR', 'NOR', 'NPL', 'NZL', 'PER', 'PHL', 'RUS', 'SAU', 'SGP', 'TUR', 'UKR', 'USA', 'ZAF']
#CAT_countries.remove('EU28')
#CAT_countries = CAT_countries + list(dt.mapp.regions.EU28.membersOf('EU28'))
dt.mapp.regions.createNewContext('political',politicallDict)
dt.mapp.regions.save()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module contains all relevant tools regarding the use of xarray


@author: ageiges
"""

# from datatoolbox import core
# import datatoolbox as dt

from time import time

from .. import core, config
import xarray as xr
import numpy as np 
import xarray as xr

import pandas as pd

#%%
def get_dimension_indices(table, dimensions):
    
    ind = list()
    for dim in dimensions:
        if isinstance(dim, tuple):
        
            index = [tuple(_get_unique_labels(table, sub_dim)[0] for sub_dim in dim)] # todo find better way
        elif dim == table.index.name:
            index = list(table.index)
        elif dim == table.columns.name:
            index = list(table.columns)
        elif dim in table.attrs.keys():
            index = [table.attrs[dim]]
        ind.append(index)
        
    return ind
    
def _get_unique_labels(table, dim):
    
    if isinstance(dim, tuple):
        unique_lables = [tuple(_get_unique_labels(table, sub_dim)[0] for sub_dim in dim)]
        # unique_lables = [tuple(d for sub_dim in dim for d in _get_unique_labels(table, sub_dim))] # todo find better way
    elif dim == table.index.name:
        unique_lables = table.index
    elif dim == table.columns.name:
        unique_lables = table.columns
    elif dim in table.meta.keys():
        unique_lables = [table.meta[dim]]
    else:
        #raise(Exception(f'Dimension {dim} not available'))
        unique_lables = [np.nan]
        
    return unique_lables

    
def get_dimensions(table_iterable, dimensions):
    
    
    dims = dict()
    
    for table in table_iterable:
    
        for dim in dimensions:
            
            dims[dim] = dims.get(dim,set()).union(_get_unique_labels(table, dim))
    return dims


#%%
    
def to_XDataSet(tableSet, dimensions=['region', 'time']):
    """
    Convert datatoolbox tableSet to a xarray data set

    Parameters
    ----------
    tableSet : datatoolbox.Tableset
        DESCRIPTION.
    dimensions : list of str
        Full dimensions of the xarray array. Other remaining dimensions will be
        added as dict like elements.

    Returns
    -------
    dSet : xarray.Dataset
        DESCRIPTION.

    """
    dimSize, dimList = core.get_dimension_extend(tableSet, dimensions= dimensions)
    
    # dimensions= ['region', 'time']
    
    dSet = xr.Dataset(coords = {key: val for (key, val) in zip(dimensions, dimList)})
    
    for key, table in tableSet.items():
        dSet[key] = table
        dSet[key].attrs = table.meta
        
    return dSet
    
def to_XDataArray(tableSet, dimensions = ['region', 'time', 'pathway']):
    #%%
#    dimensions = ['region', 'time', 'scenario', 'model']
    
#    metaDict = dict()
    
    dimSize, dimList = core.get_dimension_extend(tableSet, dimensions)
    metaCollection = core.get_meta_collection(tableSet, dimensions)
     
    xData =  xr.DataArray(np.zeros(dimSize)*np.nan, coords=dimList, dims=dimensions)
    
    for table in tableSet:
        
        indexTuple = list()
        for dim in dimensions:
            if dim == 'region':
                indexTuple.append(list(table.index))
            elif dim == 'time':
                indexTuple.append(list(table.columns))
            else:
                indexTuple.append(table.meta[dim])
                
#        xx = (table.index,table.columns,table.meta['pathway'])
        xData.loc[tuple(indexTuple)] = table.values
        
        
    # only implemented for homgeneous physical units
    assert len(metaCollection['unit']) == 1
    xData.attrs['unit'] = list(metaCollection['unit'])[0]   
    
    for  metaKey, metaValue in metaCollection.items():
        if len(metaValue) == 1:
            xData.attrs[metaKey] = metaValue.pop()
        else:
            print('Warning, dropping meta data: ' + metaKey +  ' ' + str(metaValue))
    
    return xData



def _to_xarray(tables, dimensions, stacked_dims):
    """ 
    Return a database query result as an xarray . This constuctor allows only for
    one unit, since the full array is quantified using pint-xarray. 
    The xarray dimensions (coordiantes) are defined
    by the provided dimensions. A multi-index for a coordinate can be created
    by using stacked_dims.
    
    Usage:
    -------
    tables : Iterable[[dt.Datatable]]
    dimensions :  Iterable[str]]
        Dimensions of the shared yarray dimensions / coordinates
    stacked_dims : Dict[str]]
        Dictionary of all mutli-index coordinates and their sub-dimensions
    
    Returns
    -------
    matches : xarray.Dataset + pint quantification
    """   
    tt = time()
    metaCollection = core.get_meta_collection(tables, dimensions)
    if config.DEBUG:
        print(f'ime required for meta collection: {time()-tt:2.2f}s')
    
    
    tt = time()
    final_dims = dimensions.copy()
    xdims = dimensions.copy()
    for st_dim, sub_dims in stacked_dims.items():
        [xdims.remove(dim) for dim in sub_dims]
        xdims.append(sub_dims)
        
        [final_dims.remove(dim) for dim in sub_dims]
        final_dims.append(st_dim)
        
    dims = get_dimensions(tables, xdims)
    if config.DEBUG:
        print(dims)
    coords = {x: sorted(list(dims[x])) for x in dims.keys()}
    labels = dict()
    for st_dim, sub_dims in stacked_dims.items():
        coords[st_dim] = range(len(coords[sub_dims]))
        sub_labels = list()
        for i_dim, sub_dim in enumerate(sub_dims):
            
            sub_labels.append(pd.Index([x[i_dim] for x in coords[sub_dims]], name=sub_dim))
        labels[st_dim]  = pd.MultiIndex.from_arrays(sub_labels, names = sub_dims)
        del coords[sub_dims]
    
    dimSize = [len(labels) for dim, labels in dims.items()]
    
    if config.DEBUG:
        print(f'Get timension: {time()-tt:2.2f}s')
    
    tt = time()
    xData =  xr.DataArray(np.zeros(dimSize)*np.nan, coords=coords, dims=final_dims).assign_coords(labels)
    
    tt= time()
    for table in tables:
        ind = get_dimension_indices(table,xdims)
        # xData.loc[tuple(ind)] = table.values.reshape(len(table.index),len(table.columns),1)
        
        xData.loc[tuple(ind)] = table.values.reshape(*[len(x) for x in ind])
    if config.DEBUG:
        print(f'Time required for xr data filling: {time()-tt:2.2f}s')    
    tt=time()
    metaCollection['unit'] = list(metaCollection['unit'])[0]
    xData = xData.pint.quantify(metaCollection['unit']).assign_coords(labels)
    xData.attrs = metaCollection
    
    #%%
    return xData

def key_set_to_xdataset(dict_of_data,
                        dimensions = ['model', 'scenario', 'region', 'time'],
                        stacked_dims = {'pathway': ('model', 'scenario')}):
    """ 
    Returns xarry dataset converted from a dict of pandas dataframes  or dt.TableSet. Differenty variables
    are stored as key variables. The xarray dimensions (coordiantes) are defined
    by the provided dimensions. A multi-index for a coordinate can be created
    by using stacked_dims.
    
    Usage:
    -------
    dimensions :  Iterable[str]]
        Dimensions of the shared yarray dimensions / coordinates
    stacked_dims : Dict[str]]
        Dictionary of all mutli-index coordinates and their sub-dimensions
    
    Returns
    -------
    matches : xarray.Dataset + pint quantification
    """    
    
    dim_to_sort = 'variable'
    sort_dict = dict()
    for key, table in dict_of_data.items():
        
        var = table.meta['variable']
        
        if var in sort_dict.keys():
            sort_dict[var].append(table)
        else:
            sort_dict[var] = [table]
            
    
    variables = sort_dict.keys()

    data = list()
    for variable in variables:
        tables = sort_dict[variable]
        
        xarray = _to_xarray(tables, dimensions, stacked_dims)
        data.append(xr.Dataset({variable : xarray}))
        
    ds = xr.merge(data) 

    
    return ds
    

def load_as_xdataset(query_results,
                     dimensions = ['model', 'scenario', 'region', 'time'],
                     stacked_dims = {'pathway': ('model', 'scenario')},
                     native_regions=False):
    """ 
    Returns xarry dataset pruduced from a database result. Differenty variables
    are stored as key variables. The xarray dimensions (coordiantes) are defined
    by the provided dimensions. A multi-index for a coordinate can be created
    by using stacked_dims.
    
    Usage:
    -------
    dimensions :  Iterable[str]]
        Dimensions of the shared yarray dimensions / coordinates
    stacked_dims : Dict[str]]
        Dictionary of all mutli-index coordinates and their sub-dimensions
    native_regions : bool, optional
        Load native region defintions if available. The default is False.

    Returns
    -------
    matches : xarray.Dataset + pint quantification
    """    
        
    variables = query_results.variable.unique()

    data = list()
    for variable in variables:
        tables = [core.DB.getTable(x,native_regions) for x in query_results.filterp(variable = variable).index]
        
        xarray = _to_xarray(tables, dimensions, stacked_dims)
        data.append(xr.Dataset({variable : xarray}))
        
    ds = xr.merge(data) 

    return ds

if __name__ == '__main__':
    import datatoolbox as dt
    tt = time()
    # tbs = dt.getTables(dt.find(entity = 'Emissions|CO2', source='IAMC15_2019_R2').index[:])
    
    tbs = [dt.getTable(x) for x in dt.find(entity = 'Emissions|CO2', source='IAMC15_2019_R2').index]
    print(f'Load data: {time()-tt:2.2f}s')
    dimensions   = ['model', 'scenario','median warming at peak (MAGICC6)', 'region', 'time']
    stacked_dims = {'pathway': ('model', 'scenario', 'median warming at peak (MAGICC6)')}
    xData = _to_xarray(tbs, dimensions, stacked_dims)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Mar 16 14:21:32 2020

@author: Andreas Geiges

TUTORIAL: Temperature categorization

GOAL: Classify a given variable in the IAMC database according to the temperature
      response in 2100 and visualize the outcome
      
      
"""

import datatoolbox as dt
import pandas as pd
import matplotlib.pyplot as plt

#%% SETUP

# define the data source used for this analysis
source = 'IAMC15_2019_R2'

# define the variable which is analyzed
variable = 'Emissions|CO2|Energy|Demand|Transportation'

#define years that you want to compare
years  = list(range(2000,2101,10))


# %% Data preparation
#search for the specifice reported entity 'Emissions|CO2|Energy|Demand|Transportation' within the source
searchResult = dt.find(variable =variable, source= source)

# extract the list of scenarios that report this varibble
pathways = searchResult.pathway.unique()

# dictionary to be used to save scenario/model combinations according to the temperarture response
scenarioCategory = {'1.5':[],
              '2':[],
              '3':[],
              '4':[],
              '>4':[]}

# dictionay to temporaly saving the data to be able to compute statistics
categorizedData = {'1.5':pd.DataFrame(columns=list(range(2000,2101,10))) ,
              '2':pd.DataFrame(columns=list(range(2000,2101,10))),
              '3':pd.DataFrame(columns=list(range(2000,2101,10))),
              '4':pd.DataFrame(columns=list(range(2000,2101,10))),
              '>4':pd.DataFrame(columns=list(range(2000,2101,10)))}

# looping over all scenarios that report the variable
for pathway in pathways:
    
    # ID of the temperature data for the scenarios (median of the MAGICC output)
    tableID ='__'.join(['Global_temp_MAGICC_p50|total', pathway, source])
    
    
    if dt.isAvailable(tableID): # checking that the table is in the database
        
        # reading the data using getTable
        table = dt.getTable(tableID)
        
        # dependen on the temperature in 2100, scenarios are put into different bins
        if table.loc['World',2100] < 1.5:
            scenarioCategory['1.5'].append(pathway)
        elif table.loc['World',2100] < 2:
            scenarioCategory['2'].append(pathway)
        elif table.loc['World',2100] < 3:
            scenarioCategory['3'].append(pathway)
        elif table.loc['World',2100] < 4:
            scenarioCategory['4'].append(pathway)
        else:
            scenarioCategory['>4'].append(pathway)
    else:
        pass

#%% temporarly saving the emission data 
        
for category in scenarioCategory.keys():
    
    for scenario in scenarioCategory[category]:
        tableID ='__'.join([variable, scenario, source])
        if dt.isAvailable(tableID):
            # loading the data
            table = dt.getTable(tableID) 
            
            # extracting the data for the the region "world" and selected years
            if 'World' in table.index:
                avail_years = table.columns.intersection(years)
                values = table.loc['World',avail_years] 

                #storing the values of the time series in the temporal data set
                categorizedData[category].loc[scenario] = values

          
#%% Defintion of a plotting function
def plotSubplot(subplotData, temp, scenarioCategory, color):
    
    ax = plt.subplot(*subplotData)
    for scenario in scenarioCategory[temp]:
        # creating ID
        tableID ='__'.join([variable,  scenario, source])
        if dt.isAvailable(tableID):
            # loading the data
            table = dt.getTable(tableID) 
            
            # extracting the data for the the region "world" and selected years
            if 'World' in table.index:
                avail_years = table.columns.intersection(years)
                values = table.loc['World',avail_years] 
                
                # pltting the data using matplotlib
                h = plt.plot(avail_years, values,color, linewidth=1)
        else:
            print(scenario + ' IS NOT AVAILABLE')

            
    plt.ylim([0,20000])
    plt.xlim([2000,2100])
    if subplotData[-1] > 1:
        ax.set_yticklabels([])
    plt.legend([h[0]],[ temp + '°C'])
    
#%% Create comparison plot

plt.figure(1, figsize=[10,8])

# plotting the individual scenarios
plt.clf()
plotSubplot((2,5,1), '>4', scenarioCategory, 'black')
plotSubplot((2,5,2), '4', scenarioCategory, '#DF7358')
plotSubplot((2,5,3), '3', scenarioCategory, '#E8B162')                        
plotSubplot((2,5,4), '2', scenarioCategory, '#F0E67F')     
plotSubplot((2,5,5), '1.5', scenarioCategory, '#B6CC79')     

# plotting the median of the five categories
plt.subplot(2,1,2)
plt.plot(categorizedData['>4'].columns, categorizedData['>4'].median(axis=0),'k', linewidth=2)
plt.plot(categorizedData['4'].columns, categorizedData['4'].median(axis=0),'red', linewidth=2)
plt.plot(categorizedData['3'].columns, categorizedData['3'].median(axis=0),'orange', linewidth=2)
plt.plot(categorizedData['2'].columns, categorizedData['2'].median(axis=0),'y', linewidth=2)
plt.plot(categorizedData['1.5'].columns, categorizedData['1.5'].median(axis=0),'g', linewidth=2)
plt.xlim([2000,2050])
plt.tight_layout()
plt.legend(['median ' + temp + '°C' for temp in ['>4','<4','<3','<2','<1.5']])
plt.text(2015,1150,variable)

#%%
#plt.clf()
#plotSubplot((1,1,1), '>4', scenarioCategory, 'black')
#plotSubplot((1,1,1), '4', scenarioCategory, '#DF7358')
#plotSubplot((1,1,1), '3', scenarioCategory, '#E8B162')                        
#plotSubplot((1,1,1), '2', scenarioCategory, '#F0E67F')     
#plotSubplot((1,1,1), '1.5', scenarioCategory, '#B6CC79')     
 



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu May 28 19:56:03 2020

@author: ageiges
"""
import copy
import datatoolbox as dt

from util_for_testing import df, df2, sourceMeta
 
dt.admin.switch_database_to_testing()

def test_validate():
    dt.core.DB._validateRepository()
    
def test_commit_new_table():
    df.loc['ARG', 2012] = 10
    dt.commitTable(df, 'add first table', sourceMeta)
    
    
def test_validate_ID():
    assert dt.validate_ID(list(dt.findc().index)[0])
    
def test_update_value_table():
    
    df.loc['ARG', 2012] = 20
    print(df.ID)
    dt.updateTable(df.ID, df, 'update value in table')

def test_update_meta():
    df.meta['unit'] = 'Mt CO2'
    df.meta['entity'] = 'Emissions|CO2|transport'
    oldID = copy.copy(df.ID)
    df.generateTableID()
    dt.updateTable(oldID, df, 'update meta data of table')
    
    assert  'Emissions|CO2|transport__Historic__XYZ_2020' in dt.core.DB.inventory.index
    
def test_delete_table():
    dt.removeTable('Emissions|CO2|transport__Historic__XYZ_2020')
    
    assert not dt.isAvailable(df.ID)
  
def test_commit_mutliple_tables():
    dt.commitTables([df, df2], 'adding set of table', sourceMeta)

def test_delete_mutliple_tables():
    dt.removeTables([df.ID, df2.ID])
    
    assert not dt.isAvailable(df.ID)
    assert not dt.isAvailable(df.ID)
    
def test_delete_source():    
    dt.core.DB.removeSource('XYZ_2020')
    
def test_findp():
    inv = dt.findp(variable = 'Numbers|Ones',
                   source='Numbers_2020')    
    
    
    assert len(inv.variable.unique()) ==1
    
    inv = dt.findp(variable = 'Numbers**',
                   source='Numbers_2020')    
    
    
    assert len(inv.variable.unique()) ==2
    
    
def test_table_logging():
    # create tempory table in DB
    table = dt.getTable('Numbers|Ones__Historic__Numbers_2020')
    tableNew = table*2
    tableNew.meta.update(table.meta)
    tableNew.meta.update({'category':'Twos'})
    dt.commitTable(tableNew, 'test table')
    def test_analysis():

        table = dt.getTable('Numbers|Twos__Historic__Numbers_2020')
        print(table.sum())

    # run to save all required table for the analysis
    dt.core.DB.startLogTables()
    test_analysis()
    dt.core.DB.stopLogTables()
    dt.core.DB.save_logged_tables()
    
    # remove table form database
    dt.removeTable('Numbers|Twos__Historic__Numbers_2020')
    
    #run analysis with missing table, but locally stored
    test_analysis()    
    
    #cleanup 
    import shutil
    shutil.rmtree('data')
    
if __name__ == '__main__':
    test_commit_new_table()
    test_validate_ID()
    test_update_value_table()
    test_update_meta()
    test_delete_table()
    test_commit_mutliple_tables()
    test_delete_mutliple_tables()
    test_delete_source()
    test_table_logging()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue May 11 14:10:15 2021

@author: ageiges
"""

import datatoolbox as dt
import matplotlib.pyplot as plt

re_generation = dt.getTable('Electricity_generation|Renewables__historic__IEA_WEB_2020')
total_generation = dt.getTable('Electricity_generation|Total__historic__IEA_WEB_2020')

re_share = (re_generation / total_generation).convert('%')

print(re_share)
'World' in re_share.index

re_share.loc[['World'],list(range(1990,2016,5)) + [2017, 2018,2019]].to_excel('re_share.xlsx')
re_share.meta['source'] = 'CAT calculations based one IEA WEB 2020'
re_share.loc[['World'],:].to_excel('renewables_share_CAT_2020.xlsx')

#%% solar share
new_re_generation = dt.getTable('Electricity_generation|Solar_wind_other__historic__IEA_WEB_2020')
total_generation = dt.getTable('Electricity_generation|Total__historic__IEA_WEB_2020')

new_re_share = (new_re_generation / total_generation).convert('%')

plt.clf()
plt.plot(re_share.loc['World',:], label = 'All renewables')
plt.plot(new_re_share.loc['World',:], label = 'New renewables')
plt.legend()
plt.ylabel('Share in power generation [%]')


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
----------- DATA TOOL BOX -------------
This is a python tool box project for handling global datasets. 
It contains the following features:

    Augumented pandas DataFrames adding meta data,
    Automatic unit conversion and table based computations
    ID based data structure
    Code templates (see templates.py)
    Package specific helper functions (see: tools/)

Authors: Andreas Geiges
         Jonas Hörsch     
         Gaurav Ganti
         Matthew Giddens
         
"""

from .version import version as __version__


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
try:
    # Version 0.4.5 patch of personal config
    from .settings import personal
    if not hasattr(personal, 'AUTOLOAD_SOURCES'):
        
        from .patches import patch_045_update_personal_config
        
        personal = patch_045_update_personal_config(personal)
        
    # Version path 0.4.7
    
    
except:
    pass

try:
    from .patches import patch_047_move_config_file
    patch_047_move_config_file()
except:
    pass
# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    



import os
import deprecated as _deprecated
from . import config
from . import core
from .data_structures import Datatable, TableSet, DataSet, read_csv, read_excel

try:
    from . import database
    core.DB = database.Database()
    db_connected = True
except :
    import traceback
    print('Database connection broken. Running without database connection.')    
    traceback.print_exc()
    db_connected = False
    
from . import mapping as mapp
from . import io_tools as io
from . import interfaces
from . import util as util
from . import admin as admin
from . import templates
from . import converters
# from . import utilities as _util 
#%% optional support xarray


from . import data_readers

# Predefined sets for regions and scenrarios
from datatoolbox.sets import REGIONS, SCENARIOS

import datatoolbox.tools as tools

from .tools import pandas as pd
from .tools import matplotlib as plt
from .tools import xarray as xr
from .tools import excel as xls

# unit conversion package
unitReg = core.ur

ExcelReader = xls.ExcelReader # TODO make this general IO tools

if db_connected:
    commitTable  = core.DB.commitTable
    commitTables = core.DB.commitTables
    
    updateTable             = core.DB.updateTable
    updateTables            = core.DB.updateTables
    updateTablesAvailable   = core.DB.updateTablesAvailable
    
    removeTable   = core.DB.removeTable
    removeTables  = core.DB.removeTables
    
    find                = _deprecated.deprecated(core.DB.findc, version='0.4.7', reason="Will be removed, please use finc instead")
    findc               = core.DB.findc
    findp               = core.DB.findp
    findExact           = core.DB.findExact
    getTable            = core.DB.getTable
    getTables           = core.DB.getTables
    getTablesAvailable  = core.DB.getTablesAvailable
    
    isAvailable  = core.DB._tableExists

    updateExcelInput = core.DB.updateExcelInput

insertDataIntoExcelFile = io.insertDataIntoExcelFile
# sources = _raw_sources.sources

# get country ISO code
getCountryISO = util.getCountryISO

# open_file = _util.open_file

conversionFactor = core.conversionFactor

# extract data for specific countries and sources
countryDataExtract = xls.getCountryExtract

executeForAll = util.forAll

if db_connected:
    DBinfo = core.DB.info
    sourceInfo = core.DB.sourceInfo
    inventory = core.DB.returnInventory
    
    validate_ID = core.DB.validate_ID
    #writeMAGICC6ScenFile = tools.wr
    
    # Source management
    import_new_source_from_remote = core.DB.importSourceFromRemote
    export_new_source_to_remote   = core.DB.exportSourceToRemote
    remove_source                 = core.DB.removeSource
    push_source_to_remote         = core.DB.gitManager.push_to_remote_datashelf
    pull_source_from_remote      = core.DB.pull_update_from_remote


# convenience functions
getTimeString = core.getTimeString
getDateString = core.getDateString


if db_connected:
    if config.PATH_TO_DATASHELF == os.path.join(config.MODULE_PATH, 'data/SANDBOX_datashelf'):
        print("""
              ################################################################
              You are using datatoolbox with a testing database as a SANDBOX.
              This allows for testing and initial tutorial use.
              
    
              For creating an empty dataase please use:
                  "datatoolbox.admin.create_empty_datashelf(pathToDatabase)"
    
              For switching to a existing database use: 
                  "datatoolbox.admin.change_personal_config()"
                  
                  
              ################################################################
              """)
else:
     print("""
          ################################################################
          
          You are using datatoolbox with no database connected
          
          Access functions and methods to database are not available.
              
          ################################################################
          """)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Basic data structures that allow more efficient handling of year-country
based data sets. 

Datatables are based on pandas dataframes and enfore regions as index
and years as columns. It uses meta data and provides full unit handling
by any computations using datatables.

"""

import pandas as pd
import matplotlib.pylab as plt
import numpy as np
from copy import copy
import ast
from . import core
from . import config 
import traceback
from . import util
import os
from . import tools
       
class Datatable(pd.DataFrame):
    """
    Datatable
    ^^^^^^^^^
    
    Datatable is derrived from pandas dataframe. 
    """
    _metadata = ['meta', 'ID', 'attrs']

    def __init__(self, *args, **kwargs):
        
        # pop meta out of kwargs since pandas is not expecting it
        overwrite_meta = kwargs.pop('meta', {}) # return empty dict if no meta is provide
        
        
        
        super(Datatable, self).__init__(*args, **kwargs)
        
        if len(args) > 0 and hasattr(args[0], 'meta'):
            # print(args)
            meta = args[0].meta
            
        elif 'data' in kwargs.keys() and hasattr(kwargs['data'], 'meta'):
            meta = kwargs['data'].meta
            
        else:
            meta = {}
        
        meta.update(overwrite_meta)

        # append metaDict to datatable
        self.__appendMetaData__(meta)
        
        # self.vis = Visualization(self)
        try:
            self.generateTableID()
        except:
            self.ID = None
    
        if config.AVAILABLE_XARRAY:
            self.to_xarray = self._to_xarray
            
        self.columns.name = config.DATATABLE_COLUMN_NAME
        self.index.name   = config.DATATABLE_INDEX_NAME
        
        #print(self.meta)
        if ('_timeformat' in self.meta.keys()) and (self.meta['_timeformat'] != '%Y'
                                                    and self.meta['_timeformat'] != 'Y'):
            self.columns_to_datetime()

        self.attrs = self.meta
    
    def info(self):
        """
        Returns information about the dataframe like shape, index and column 
        extend and the number of non-nan entries.
        
        Returns
        -------
        str
            Information about datatable.

        """
        shp = self.shape
        if (len(self.index) == 0) or (len(self.columns) == 0):
            return ('Empty Datatable')
        idx_ext = self.index[0] ,self.index[-1]
        time_ext = self.columns[0] ,self.columns[-1]
        n_entries = (~self.isnull()).sum().sum()
        return f'{shp[0]}x{shp[1]} Datatable with {n_entries} entries, index from {idx_ext[0]} to {idx_ext[1]} and time from {time_ext[0]} to {time_ext[0]}'
        
        
    @classmethod
    def from_pyam(cls, idf, **kwargs):
        """
        Create a datatable from an iam dataframe.

        Parameters
        ----------
        cls : datatable class
            DESCRIPTION.
        idf : pyam dataframe
            dataframe that contrains the data that is used to create the datatable.
        **kwargs : TYPE
            DESCRIPTION.

        Returns
        -------
        datatatble : datatoolbox datatable
            Datatable with original unit and related meta data.

        """
        import pyam

        if kwargs:
            idf = idf.filter(**kwargs)

        assert len(idf.variable) == 1, (
            f"Datatables cannot represent more than one variable, "
            f"but there are {', '.join(idf.variable)}"
        )

        def extract_unique_values(df, fields, ignore):
            meta = {}
            for fld in set(fields).difference(ignore):
                values = df[fld].unique()
                assert len(values) == 1, (
                    f"Datatables can only represent unique meta entries, "
                    f"but {fld} has {', '.join(values)}"
                )
                meta[fld] = values[0]
            return meta

        meta = {
            **extract_unique_values(idf.data, pyam.IAMC_IDX, ['region']),
            **extract_unique_values(idf.meta, idf.meta.columns, ['exclude'])
        }

        data = (
            idf.data.pivot_table(index=['region'], columns=idf.time_col)
            .value  # column name
            .rename_axis(columns=None)
        )

        return cls(data, meta=meta)

    @classmethod
    def from_excel(cls, filepath, sheetName = None):
        """
        Create a dataframe from a suitable excel file that is saved by datatoolbox.
        
        Parameters
        ----------
        cls : class
        
        filepath : str
            Path to the file.
        sheetName : str, optional
            Sheetn ame that is read in. The default is None.
     
        Returns
        -------
        datatable
            DESCRIPTION.
        """
        if sheetName is None:
            sheetNames = None
        else:
            sheetNames= [sheetNames]
        return read_excel(filepath,
                          sheetNames=sheetNames)
       
    def _to_xarray(self):
        
        return core.xr.DataArray(self.values, 
                                 coords=[self.index, self.columns], 
                                 dims=['space','time'], 
                                 attrs=self.meta)
    

    
    @property
    def _constructor(self):
        return Datatable
    
    def __finalize__(self, other, method=None, **kwargs):
        """propagate metadata from other to self """
        # merge operation: using metadata of the left object
        if method == 'merge':
            for name in self._metadata:
                object.__setattr__(self, name, copy(getattr(other.left, name, None)))
        # concat operation: using metadata of the first object
        elif method == 'concat':
            for name in self._metadata:
                object.__setattr__(self, name, copy(getattr(other.objs[0], name, None)))
        else:
            for name in self._metadata:
                #print(other)
                object.__setattr__(self, name, copy(getattr(other, name, None)))
        return self    
    
   
    def __appendMetaData__(self, metaDict):    
        """
        Private function to append meta data.

        Parameters
        ----------
        metaDict : dict
            New meta data.


        """
        
        for metaKey in config.REQUIRED_META_FIELDS:
            if metaKey not in metaDict.keys() or metaDict[metaKey] == '' or pd.isna(metaDict[metaKey]):
                
                # Overwrite with default or empty string
                if metaKey in config.META_DEFAULTS:
                    metaDict[metaKey] = config.META_DEFAULTS[metaKey]
                else:
                    metaDict[metaKey] = ''

        
        self.__setattr__('meta', metaDict.copy())
        
        assert core.getUnit(self.meta['unit'])

    def columns_to_datetime(self):
        self.columns = pd.to_datetime(self.columns, 
                                      format=self.meta['_timeformat'])

    def copy(self, deep=True):
        """
        Make a copy of this Datatable object
        Parameters
        ----------
        deep : boolean, default True
            Make a deep copy, i.e. also copy data
        Returns
        -------
        copy : Datatable
        """
        # FIXME: this will likely be unnecessary in pandas >= 0.13
        data = self._data
        if deep:
            data = data.copy(deep=True)
        return Datatable(data).__finalize__(self) 

    def diff(self, 
             periods=1, 
             axis=0):
        """
        Compute the difference between different years in the datatable
        Equivalent do pandas diff but return datatable.
        
        Parameters
        ----------
        periods : int, optional
            DESCRIPTION. The default is 1.
        axis : int, optional
            DESCRIPTION. The default is 0.

        Returns
        -------
        out : TYPE
            DESCRIPTION.

        """
        out = super(Datatable, self).diff(periods=periods, axis=axis)
        out.meta['unit'] = self.meta['unit']
        
        return out
    
    def reduce(self, 
               method='linear_piece_wise',
               eps = 1e-6):
        """ 
        Reduce data that is piecewise linear to the core data points (kinks).
        """
        
        # assert monotonic incease of decrease
        assert self.columns.is_monotonic
        
        #initial value of last year (set to first year)
        last_year = self.columns[0]
        last_yearly_change    = np.nan # initial value
        
        reduced_data = self.copy()*np.nan
        
        for year in self.columns[1:]:
            #assert year == last_year+1
            n_years = year - last_year
            
            change= self.loc[:,year] - self.loc[:,last_year]
            
            # find where is a kink
            idx = (change - last_yearly_change*n_years).abs() > eps
            
            reduced_data.loc[idx, last_year] = self.loc[idx,last_year]
            
            # overwrite last values
            last_yearly_change = change / n_years
            last_year = year
        
        # set first and last values
        for coISO in self.index:
            value_columns = self.columns[self.loc[coISO,:].notna()]
            idx_min = value_columns.min()
            idx_max = value_columns.max()
            reduced_data.loc[coISO, idx_min] = self.loc[coISO,idx_min]
            reduced_data.loc[coISO, idx_max] = self.loc[coISO,idx_max]
            
            #%%
        return reduced_data
    
    def to_excel(self, fileName = None, sheetName = "Sheet0", writer = None, append=False):
        """
        Save datatable to excel.

        Parameters
        ----------
        fileName : str, optional
            Relative file path. If None is provide, a writer is expected. The default is None.
        sheetName : str, optional
            Sheet name that is read in. The default is "Sheet0".
        writer : pandas excel writer, optional
            Pandas writer that is used instead opening a new one. The default is None.
        append : bool, optional
            If true, data is appended to the writer. The default is False.

        Returns
        -------
        None.

        """

        if fileName is not None:
            if append:
                writer = pd.ExcelWriter(fileName, 
                                        engine='openpyxl', mode='a',
                                        datetime_format='mmm d yyyy hh:mm:ss',
                                        date_format='mmmm dd yyyy')  
            else:
                writer = pd.ExcelWriter(fileName,
                                        engine='xlsxwriter',
                                        datetime_format='mmm d yyyy hh:mm:ss',
                                        date_format='mmmm dd yyyy')  
            
        metaSeries= pd.Series(data=[''] + list(self.meta.values()) + [''],
                              index=['###META###'] + list(self.meta.keys()) + ['###DATA###'])
        
        metaSeries.to_excel(writer, sheet_name=sheetName, header=None, columns=None)
        super(Datatable, self).to_excel(writer, sheet_name= sheetName, startrow=len(metaSeries))
        
        if fileName is not None:
            writer.close()
        
        
        
    
    def to_csv(self, fileName=None):
        """
        Save the datatable to an annotated csv file.

        Parameters
        ----------
        fileName : str, optional
            Path to file. The default is None.

        Returns
        -------
        None.

        """
        if fileName is None:
            fileName = '|'.join([ self.meta[key] for key in config.ID_FIELDS]) + '.csv'
        else:
            assert fileName[-4:]  == '.csv'
        
#         fid = open(fileName,'w', encoding='utf-8')
#         fid.write(config.META_DECLARATION)
        
#         for key, value in sorted(self.meta.items()):
# #            if key == 'unit':
# #                value = str(value.u)
#             fid.write(key + ',' + str(value) + '\n')
        
#         fid.write(config.DATA_DECLARATION)
#         super(Datatable, self).to_csv(fid)
#         fid.close()
        
        core.csv_writer(fileName,
                        pd.DataFrame(self),
                        self.meta)

    def to_pyam(self, **kwargs):
        """
        Conversion to pyam dataframe.

        Parameters
        ----------
        **kwargs : TYPE
            DESCRIPTION.

        Raises
        ------
        AssertionError
            DESCRIPTION.

        Returns
        -------
        idf : pyam dataframe
            DESCRIPTION.

        """
        from pyam import IamDataFrame

        meta = {
            **self.meta,
            **kwargs
        }

        try:
            idf = IamDataFrame(
                pd.DataFrame(self).rename_axis(index="region").reset_index(),
                model=meta.get('model', ''),
                scenario=meta["scenario"],
                variable=meta['variable'],
                unit=meta['unit']
            )
        except KeyError as exc:
            raise AssertionError(f"meta does not contain {exc.args[0]}")

        # Add model, scenario meta fields
        for field in ('pathway', 'source', 'source_name', 'source_year'):
            if field in meta:
                idf.set_meta(meta[field], field)
 
        return idf
       
    def to_IamDataFrame(self, **kwargs):
        """
        Function to sustain backwars compatibility
        Depreciated.

        """
        return self.to_pyam(**kwargs)
        
    def convert(self, newUnit, context=None):
        """
        Convert datatable to different unit and returns converted
        datatable.

        Parameters
        ----------
        newUnit : str
            New unit string in which the datatable should be converted.
        context : str, optional
            Optional context (e.g. GWPAR4). The default is None.

        Returns
        -------
        datatable
            Datatable converted in the new unit.

        """
        if self.meta['unit'] == newUnit:
            return self
        
        dfNew = self.copy()
#        oldUnit = core.getUnit(self.meta['unit'])
#        factor = (1* oldUnit).to(newUnit).m
        
        factor = core.conversionFactor(self.meta['unit'], newUnit, context)
        
        dfNew.loc[:] =self.values * factor
        dfNew.meta['unit'] = newUnit
        dfNew.meta['modified'] = core.getTimeString()
        return dfNew
        
    def aggregate_region(self, mapping, skipna=False):
        """ 
        This functions added the aggregates to the table according to the provided
        mapping.( See datatools.mapp.regions)
        
        Returns the result, but does not inplace add it.
        """
        from datatoolbox.tools.for_datatables import aggregate_region
        return aggregate_region(self, mapping, skipna)
    
    def interpolate(self, method="linear", add_missing_years=False):
        """
        Interpoltate missing data between year with the option to add
        missing years in the columns.

        Parameters
        ----------
        method : sting, optional
            Interpolation method. The default is "linear".
            - linear
        add_missing_years : bool, optional
            If true, missing years within the time value range are added to 
            the dataframe. The default is False.

        Returns
        -------
        datatable
            Interpolated dataframe.

        """
        from datatoolbox.tools.for_datatables import interpolate
        
        if add_missing_years:
            for col in list(range(self.columns.min(),self.columns.max()+1)):
                if col not in self.columns:
                    self.loc[:,col] = np.nan
            self = self.loc[:,list(range(self.columns.min(),self.columns.max()+1))]
        
        return interpolate(self, method)
    
    def clean(self):
        """
        Clean up the dataframe to only recogniszed regions, years and numeric values. 
        Removed columns and rows with only nan values.

        Returns
        -------
        datatable
            DESCRIPTION.

        """
        return util.cleanDataTable(self)
    
    def filter(self, spaceIDs):
        """
        Filter dataframe based on a list of spatial IDs.

        Parameters
        ----------
        spaceIDs : iterable of str
            DESCRIPTION.

        Returns
        -------
        datatable
            DESCRIPTION.

        """
        mask = self.index.isin(spaceIDs)
        return self.iloc[mask,:]
    
    
    def yearlyChange(self,forward=True):
        """
        This methods returns the yearly change for all years (t1) that reported
        and and where the previous year (t0) is also reported
        
        Parameters
        ----------
        forward : bool 
            If true, the yearly change is computed in the forward direction, otherwise
            backwards.
            Default is forward.
        """

        #%%
        if forward:
            t0_years = self.columns[:-1]
            t1_years = self.columns[1:]
            index    = self.index
            t1_data  = self.iloc[:,1:].values
            t0_data  = self.iloc[:,:-1].values
            
            deltaData = Datatable(index=index, columns=t0_years, meta={key:self.meta[key] for key in config.REQUIRED_META_FIELDS})
            deltaData.meta['entity'] = 'delta_' + deltaData.meta['entity']
            deltaData.loc[:,:] = t1_data - t0_data
        else:
                
            t1_years = self.columns[1:]
            index    = self.index
            t1_data  = self.iloc[:,1:].values
            t0_data  = self.iloc[:,:-1].values
            
            deltaData = Datatable(index=index, columns=t1_years, meta={key:self.meta[key] for key in config.REQUIRED_META_FIELDS})
            deltaData.meta['entity'] = 'delta_' + deltaData.meta['entity']
            deltaData.loc[:,:] = t1_data - t0_data
        
        
        return deltaData
    
    #%%
    def generateTableID(self):
        """
        Generates the table ID based on the meta data of the table.

        Returns
        -------
        datatable
            DESCRIPTION.

        """
        # update meta data required for the ID
        self.meta =  core._update_meta(self.meta)
        self.ID   =  core._createDatabaseID(self.meta)
        self.meta['ID'] = self.ID
        return self.ID
    
    def getTableFilePath(self):
        source = self.meta['source']
        fileName = core.generate_table_file_name(self.ID)
        return os.path.join(config.PATH_TO_DATASHELF, 'database/', source, 'tables', fileName)

        
    def getTableFileName(self):
        """
        For compatibility to windows based sytems, the pipe '|' symbols is replaces
        by double underscore '__' for the csv filename.
        """
        self.generateTableID()
        
        return core.generate_table_file_name(self.ID)
    
    def _update_meta(self):
        self.meta =  core._update_meta(self.meta)
    
    def source(self):
        """
        Return the source of the table 
        """
        return self.meta['source']

    def append(self, other, **kwargs):
        """
        Append data to the datatable
        

        Parameters
        ----------
        other : datatable
            New data that will be added to the datatable.
        **kwargs : TYPE
            Default pandas append arguments.

        Returns
        -------
        datatable

        """
        kwargs.setdefault("sort", True)

        if isinstance(other,Datatable):
            
            if other.meta['entity'] != self.meta['entity']:
#                print(other.meta['entity'] )
#                print(self.meta['entity'])
                raise(BaseException('Physical entities do not match, please correct'))
            if other.meta['unit'] != self.meta['unit']:
                other = other.convert(self.meta['unit'])
        
        out =  pd.concat([self, other], **kwargs)
        
        # only copy required keys
        out.meta = {key: value for key, value in self.meta.items() if key in config.REQUIRED_META_FIELDS}
        
        # overwrite scenario
        out.meta['scenario'] = 'computed: ' + self.meta['scenario'] + '+' + other.meta['scenario']
        return out

    
        
    def __add__(self, other):
        """
        Private function to add two dataframes. The added table is converted to
        the unit of first table.

        Parameters
        ----------
        other : datatble
            Data to add.

        Returns
        -------
        out : datatable
            DESCRIPTION.

        """
        if isinstance(other,Datatable):
            
            if self.meta['unit'] == other.meta['unit']:
                factor = 1
            else:
                factor = core.getUnit(other.meta['unit']).to(self.meta['unit']).m
            
            rhs = pd.DataFrame(other * factor)
            out = Datatable(super(Datatable, self.copy()).__add__(rhs))

            out.meta['unit'] = self.meta['unit']
            out.meta['source'] = 'calculation'
        else:
            out = Datatable(super(Datatable, self).__add__(other))
            out.meta['unit'] = self.meta['unit']
            out.meta['source'] = 'calculation'
        return out 

    __radd__ = __add__
    
    def __sub__(self, other):
        """
        Private function to subract two dataframes. The subracted table is converted to
        the unit of first table.

        Parameters
        ----------
        other : datatble
            Data to substract.

        Returns
        -------
        out : datatable
            DESCRIPTION.

        """
        if isinstance(other,Datatable):
            if self.meta['unit'] == other.meta['unit']:
                factor = 1
            else:
                factor = core.getUnit(other.meta['unit']).to(self.meta['unit']).m
            rhs = pd.DataFrame(other * factor)
            out = Datatable(super(Datatable, self).__sub__(rhs))
            out.meta['unit'] = self.meta['unit']
            out.meta['source'] = 'calculation'
        else:
            out = Datatable(super(Datatable, self).__sub__(other))
            out.meta['unit'] = self.meta['unit']
            out.meta['source'] = 'calculation'
        return out
    
    def __rsub__(self, other):
        """
        Equivalent to __sub__
        """
        if isinstance(other,Datatable):
            if self.meta['unit'] == other.meta['unit']:
                factor = 1
            else:
                factor = core.getUnit(other.meta['unit']).to(self.meta['unit']).m
            out = Datatable(super(Datatable, self).__rsub__(other * factor))
            out.meta['unit'] = self.meta['unit']
            out.meta['source'] = 'calculation'
        else:
            out = Datatable(super(Datatable, self).__rsub__(other))
            out.meta['unit'] = self.meta['unit']
            out.meta['source'] = 'calculation'
        return out
        
    def __mul__(self, other):
        if isinstance(other,Datatable):
            newUnit = (core.getUnit(self.meta['unit']) * core.getUnit(other.meta['unit']))
            out = Datatable(super(Datatable, self).__mul__(other))
            out.meta['unit'] = str(newUnit.u)
            out.meta['source'] = 'calculation'
            out.values[:] *= newUnit.m
        else:
            out = Datatable(super(Datatable, self).__mul__(other))
            out.meta['unit'] = self.meta['unit']
            out.meta['source'] = 'calculation'
        return out    
    
    __rmul__ = __mul__

    def __truediv__(self, other):
        if isinstance(other,Datatable):
            newUnit = (core.getUnit(self.meta['unit']) / core.getUnit(other.meta['unit']))
            out = Datatable(super(Datatable, self).__truediv__(other))
            out.meta['unit'] = str(newUnit.u)
            out.meta['source'] = 'calculation'
            out.values[:] *= newUnit.m
        else:
            out = Datatable(super(Datatable, self).__truediv__(other))
            out.meta['unit'] = self.meta['unit']
            out.meta['source'] = 'calculation'
        return out

#    __rtruediv__ = __truediv__
    def __rtruediv__(self, other):
        if isinstance(other,Datatable):
            newUnit = (core.getUnit(other.meta['unit']) / core.getUnit(self.meta['unit']))
            out = Datatable(super(Datatable, self).__rtruediv__(other))
            out.meta['unit'] = str(newUnit.u)
            out.meta['source'] = 'calculation'
            out.values[:] *= newUnit.m
        else:
            out = Datatable(super(Datatable, self).__rtruediv__(other))
            out.meta['unit'] = str((core.getUnit(self.meta['unit'])**-1).u)
            out.meta['source'] = 'calculation'
        return out
    
    def __repr__(self):
        outStr = """"""
        if 'ID' in self.meta.keys():
            outStr += '=== Datatable - ' + self.meta['ID'] + ' ===\n'
        else:
            outStr += '=== Datatable ===\n'
        for key in self.meta.keys():
            if self.meta[key] is not None and not str(self.meta[key]).startswith('_'):
                outStr += key + ': ' + str(self.meta[key]) + ' \n'
        outStr += super(Datatable, self).__repr__()
        return outStr
    
    def _repr_html_(self):
        outStr = """"""
        if 'ID' in self.meta.keys():
            outStr += '=== Datatable - ' + self.meta['ID'] + ' ===<br/>\n'
        else:
            outStr += '=== Datatable ===<br/>\n'
        for key in self.meta.keys():
            if self.meta[key] is not None and not str(self.meta[key]).startswith('_'):
                outStr += key + ': ' + str(self.meta[key]) + ' <br/>\n'
        outStr += super(Datatable, self)._repr_html_()
        return outStr
#%%
class TableSet(dict):
    """
    Class TableSet that is inherited from the dict class. It organized multiple 
    heterogeneous datatbles into one structure.
    """
    def __init__(self, IDList=None):
        """
        Create tableset from a given list of table IDs. All tables are loaded from
        the database.

        Parameters
        ----------
        IDList : list, optional
            DESCRIPTION. The default is None.

        Returns
        -------
        None.

        """
        super(dict, self).__init__()
        self.inventory = pd.DataFrame(columns = ['key']+ config.INVENTORY_FIELDS)
        
        if IDList is not None:
            for tableID in IDList:
                self.add(core.DB.getTable(tableID))
    
    def __getitem__(self, key):
        item = super(TableSet, self).__getitem__(key)
        
        #load datatable if necessary
        if item is None:
            item = core.DB.getTable(key)
            self[key] = item
        
        return item

    def __setitem__(self, key, datatable):
        super(TableSet, self).__setitem__(key, datatable)
        
        if datatable.ID is None:
            try:
                datatable.generateTableID()
            except:
#                print('Could not generate ID, key used instead')
                datatable.ID = key
        self.inventory.loc[datatable.ID] = None
        self.inventory.loc[datatable.ID, "key"] = key
        self.inventory.loc[datatable.ID, config.INVENTORY_FIELDS] = [datatable.meta.get(x,None) for x in config.INVENTORY_FIELDS]
    
        
        

    

    def __iter__(self):
        return iter(self.values())
    

            
    def _add_list(self, tableList):
        for table in tableList:
            tableID = table.generateTableID()
            
            if tableID in self.keys():
                self._update(table, tableID)
            else:
                self.__setitem__(tableID, table)
    
    def _add_TableSet(self, tableSet):
        
        for tableID, table in tableSet.items():
            
            if tableID in self.keys():
                self._update(table, tableID)
            else:
                self.__setitem__(tableID, table)
    
    def _add_single_table(self, table):
        tableID = table.generateTableID()
        if tableID in self.keys():
            self._update(table, tableID)
        else:
            self.__setitem__(tableID, table)
        
    def _add_tableID(self, tableID):
        self[tableID] = None
        self.inventory.loc[tableID] = [None for x in config.ID_FIELDS]   
        
        
    def _update(self, table, tableKey):
        
        # make sure the data is compatible
#        print(table.meta)
#        print(self[tableKey].meta)
        if table.meta != self[tableKey].meta:
            raise(BaseException('Trying to update table with different meta'))
        
        # update data
        self[tableKey] = pd.concat([self[tableKey] , table])
            
    def add(self, datatables=None, tableID=None):
        """
        Add new tables to table set. Either datatables or table IDs should be given.

        Parameters
        ----------
        datatables : list of datatables, optional
            DESCRIPTION. The default is None.
        tableID : TYPE, optional
            DESCRIPTION. The default is None.

        Returns
        -------
        None.

        """
        
        # assert only on parameter is None
        assert not ((datatables is None) and (tableID is None))
        
        if datatables is not None:
            
            if isinstance(datatables, list):
                self._add_list(datatables)
                
            elif isinstance(datatables, TableSet):
                self._add_TableSet(datatables)
                
            elif isinstance(datatables, Datatable):
                self._add_single_table(datatables)
                
            else:
                 print('Data type not recognized.')
            
        elif tableID is not None:
            self._add_tableID(tableID)
          
    
    def convert(self, newUnit):
        """
        Convert all tables to a new Unit. Returns a copy of the old tableSet

        Parameters
        ----------
        newUnit : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        tables = self.copy()
        
        for tableKey in tables.keys():
            
            tables[tableKey].convert(newUnit)
            tables[tableKey].meta['unit'] = newUnit
            tables.inventory.loc[tableKey,'unit'] = newUnit
            
        return tables
    
    def copy(self):
        return copy(self)
    
    def remove(self, tableID):
        """
        Remove table form tableSet.

        Parameters
        ----------
        tableID : str
            TableID.

        Returns
        -------
        None.

        """
        del self[tableID]
        self.inventory.drop(tableID, inplace=True)
        
    def filterp(self, level=None, regex=False, **filters):
        """ 
        Future defaulf find method that allows for more
        sophisticated syntax in the filtering
        
        Usage:
        -------
        filters : Union[str, Iterable[str]]
            One or multiple patterns, which are OR'd together
        regex : bool, optional
            Accept plain regex syntax instead of shell-style, default: False
        
        Returns
        -------
        matches : pd.Series
        Mask for selecting matched rows
        """    
            
        # filter by columns and list of values
        keep = True

        for field, pattern in filters.items():
            # treat `col=None` as no filter applied
            if pattern is None:
                continue

            if field not in self.inventory:
                raise ValueError(f'filter by `{field}` not supported')

            keep &= util.pattern_match(
                self.inventory[field], pattern, regex=regex
            )

        if level is not None:
            keep &= self.inventory['variable'].str.count(r"\|") == level

        return self.inventory if keep is True else self.inventory.loc[keep]
            
    
    def filter(self,**kwargs):
        """
        Filter tableSet based on the given table inventory columns. 
        (see config.INVENTORY_FIELDS)

        Parameters
        ----------
        **kwargs : dict-like
            Filter arguments as string that need to be contained in the fields..

        Returns
        -------
        newTableSet : tableSet
            DESCRIPTION.

        """
        inv = self.inventory.copy()
        for key in kwargs.keys():
            #table = table.loc[self.inventory[key] == kwargs[key]]
            mask = self.inventory[key].str.contains(kwargs[key], regex=False)
            mask[pd.isna(mask)] = False
            mask = mask.astype(bool)
            inv = inv.loc[mask].copy()
            
        newTableSet = TableSet()
        for key in inv.index:
            newTableSet[key] = self[key]
            
        return newTableSet

    @classmethod
    def from_list(cls,
                  tableList):
        """
        Create tableSet form list of datatables.

        Parameters
        ----------
        cls : TYPE
            DESCRIPTION.
        tableList : list
            List of datatables.

        Returns
        -------
        tableSet : tableset
            DESCRIPTION.

        """
        tableSet = cls()
        for table in tableList:
            tableSet.add(table)
    
        return tableSet
    
    def aggregate_to_region(self, mapping):
        """ 
        This functions added the aggregates to the output according to the provided
        mapping.( See datatools.mapp.regions)
        
        Returns the result, but does not inplace add it.
        """
        return util.aggregate_tableset_to_region(self, mapping)
        
        
    
    #%%
        
    def to_compact_excel(self,
                         writer,
                         sheet_name="Sheet1",
                         include_id=False,
                         meta_columns = None):
        
        use_index = include_id

        if isinstance(writer, pd.ExcelWriter):
            need_close = False
        else:
            writer = pd.ExcelWriter(pd.io.common.stringify_path(writer))
            need_close = True
            
        
        long, metaDict = self.to_compact_long_format(include_id, meta_columns=meta_columns)
    
        if meta_columns is not None:
            for metaKey in meta_columns:
                if metaKey not in long.columns:
                    long.loc[:,metaKey] = metaDict[metaKey]
            years = util.yearsColumnsOnly(long)
            long = long.loc[:, meta_columns + years]

    
        core.excel_writer(writer,
                         long,
                         metaDict,
                         sheet_name=sheet_name,
                         index= use_index,
                         engine='xlsxwriter')
        
        if need_close:
            writer.close()
            
    
    def to_excel(self, 
                 fileName, 
                 append=False,
                 compact = False):
        """
        Sace TableSet as excel file with individual datatables in individual sheets.
    
        Parameters
        ----------
        fileName : str
            File path.
        append : bool, optional
            If true, try to append data. The default is False.
    
        Returns
        -------
        None.

        """

        if append:
            writer = pd.ExcelWriter(fileName, 
                                    engine='openpyxl', mode='a',
                                    datetime_format='mmm d yyyy hh:mm:ss',
                                    date_format='mmmm dd yyyy')  
        else:
            writer = pd.ExcelWriter(fileName,
                                    engine='xlsxwriter',
                                    datetime_format='mmm d yyyy hh:mm:ss',
                                    date_format='mmmm dd yyyy')  
        

            
        for i,eKey in enumerate(self.keys()):
            table = self[eKey].dropna(how='all', axis=1).dropna(how='all', axis=0)
            sheetName = str(i) + table.meta['ID'][:25]
#            print(sheetName)
            table.to_excel(writer=writer, sheetName = sheetName)
            
        writer.close()
        
    def create_country_dataframes(self, countryList=None, timeIdxList= None):
        
        # using first table to get country list
        if countryList is None:
            countryList = self[list(self.keys())[0]].index
        
        coTables = dict()
        
        for country in countryList:
            coTables[country] = pd.DataFrame([], columns= ['entity', 'unit', 'source'] +list(range(1500,2100)))
            
            for eKey in self.keys():
                table = self[eKey]
                if country in table.index:
                    coTables[country].loc[eKey,:] = table.loc[country]
                else:
                    coTables[country].loc[eKey,:] = np.nan
                coTables[country].loc[eKey,'source'] = table.meta['source']
                coTables[country].loc[eKey,'unit'] = table.meta['unit']
                                    
            coTables[country] = coTables[country].dropna(axis=1, how='all')
            
            if timeIdxList is not None:
                
                containedList = [x for x in timeIdxList if x in coTables[country].columns]
                coTables[country] = coTables[country][['source', 'unit'] + containedList]

            
        return coTables

    def variables(self):
        return list(self.inventory.variable.unique())

    def pathways(self):
        return list(self.inventory.pathway.unique())

    def entities(self):
        return list(self.inventory.entity.unique())

    def scenarios(self):
        return list(self.inventory.scenario.unique())
    
    def sources(self):
        return list(self.inventory.source.unique())

    def sum(self, new_meta= {}):
        """
        This will sum up all tables in the tableSet if units do allow it. The user
        needs to make sure that the computation does make sense. 
        
        If meta data is provided, the new table will be updated using this meta

        Returns
        -------
        resultTable : TYPE
            DESCRIPTION.

        """
        keyList = list(self.keys())
        
        #copy first element
        resultTable = self[keyList[0]].copy()
        
        # loop over remainen elements
        for key in keyList[1:]:
            resultTable  =resultTable +  self[key]
         
            
        #updating with new meta data if provided
        resultTable.meta.update(new_meta)
            
        return resultTable

    def get_all_meta_keys(self):
        meta_columns = set()
        for key in self.keys():
            meta = self[key].meta
            
            meta_columns = meta_columns.union(meta.keys())
        return list(meta_columns)
    
    
    def to_compact_long_format(self, 
                                include_id= False,
                                meta_columns=None):
        
        if meta_columns is None:
            meta_columns = ['region'] + self.get_all_meta_keys()
        # meta_columns = ['region'] + config.INVENTORY_FIELDS
        if include_id:
            meta_columns.append('ID')
        
                
        
        long = self.to_LongTable(meta_list = meta_columns)
        
        single_meta = dict()
        multi_meta  = list()
        columns_to_drop = list()
        
        for column in meta_columns:
            unique_entries = long.loc[:,column].unique()
            if len(unique_entries) == 1:
                if unique_entries[0] != '':
                    single_meta[column] = unique_entries[0]
                
                columns_to_drop.append(column)
            else:
                multi_meta.append(column)
        long = long.drop(columns_to_drop, axis = 1)
        
        metaDict = dict()
        metaDict.update(single_meta)
        metaDict.update({'meta_columns' : multi_meta})
        
        if include_id:
            long = long.set_index('ID')
        
        return long, metaDict

    def to_csv(self,
               filename):
        """
        Conversion to compact lone csv format

        Parameters
        ----------
        filename : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        #%%

        long, metaDict = self._compact_to_long_format()
        
        core.csv_writer(filename, 
                        long,
                        metaDict,
                        index=None)
    #%%           
    def to_xarray(self, dimensions=None):
        """
        Convert tableset to and xarray with the given dimenions.
        Requires xarray installed

        Parameters
        ----------
        dimensions : list of str
            List of xarray dimensions.

        Returns
        -------
        xr.xarray
            DESCRIPTION.

        """
        if dimensions is None:
            dimensions = ['region', 'time']
            for col in config.ID_FIELDS + ['unit']:
                if len(self.inventory.loc[:,col].unique()) > 1:
                    dimensions.append(col)
                    
        if "unit" in dimensions:
            raise(BaseException('Different units in dataset can not be merged in one xarray'))
        
        if not config.AVAILABLE_XARRAY:
            raise(BaseException('module xarray not available'))
        return tools.xarray.to_XDataArray(self, dimensions)
       
    def to_xset(self, dimensions = ['region', 'time']):
        """
        Convert table set to an xarray data set.

        Parameters
        ----------
        dimensions : list, optional
            DESCRIPTION. The default is ['region', 'time'].

        Returns
        -------
        xr.Dataset
            DESCRIPTION.

        """
        dimensions = ['region', 'time']
        if not config.AVAILABLE_XARRAY:
            raise(BaseException('module xarray not available'))
        return tools.xarray.to_XDataSet(self, dimensions)
    
    def to_list(self):
        """
        Convert to list of tables

        Returns
        -------
        list
            List of datatables.

        """
        return [ self[key] for key in self.keys()]
    
        
    def to_LongTable(self, 
                     native_regions=False,
                     meta_list = ['variable', 'region','scenario', 'model', 'unit']):
        tables = []

        for variable, df in self.items():
            if df.empty:
                continue
            inp_dict = dict()    
            
            if isinstance(df.index, pd.MultiIndex):
                if native_regions:
                    df = df.reset_index('standard_region',drop=True)
                else:
                    df = df.reset_index('region',drop=True)
                
            for metaKey in meta_list:
                if metaKey == 'region':
                    inp_dict[metaKey] = df.index
                else:
                    inp_dict[metaKey] = df.meta.get(metaKey, '')
                
            try:
                df = pd.DataFrame(df.assign(**inp_dict)).reset_index(drop=True)
                
            except KeyError as exc:
                raise AssertionError(f"meta of {variable} does not contain {exc.args[0]}")
 
            tables.append(df)

        long_df = pd.concat(tables, ignore_index=True, sort=False)
        
        # move id columns to the front
        id_cols = pd.Index(meta_list)
        long_df = long_df[list(id_cols) + list( long_df.columns.difference(id_cols))]
        long_df = pd.DataFrame(long_df)
        return long_df

    def to_pyam(self):
        
        import pyam
        long_table = self.to_LongTable()
        long_table.index.name = None
        
        # make sure that region does not contain any nan values
        na_mask = long_table['region'].isnull()
        if config.DEBUG and (sum(na_mask) >0 ):
            print(f'Removing {sum(na_mask)} nan items in region index')
        long_table = long_table[~na_mask]
        
        idf = pyam.IamDataFrame(pd.DataFrame(long_table))

        meta = pd.DataFrame([df.meta for df in self.values()])
        if 'model' not in meta:
            meta['model'] = ""
        if 'scenario' not in meta:
            meta['scenario'] = ""
        meta = (
            meta[
                pd.Index(['model', 'scenario', 'pathway'])
                .append(meta.columns[meta.columns.str.startswith('source')])
            ]
            .set_index(['model', 'scenario'])
            .drop_duplicates()
        )

        idf.meta = meta
        idf.reset_exclude()

        return idf

    # Alias for backwards-compatibility
    to_IamDataFrame = to_pyam

    def plotAvailibility(self, regionList= None, years = None):
        
        avail= 0
        for table in self:
#            print(table.ID)
            table.meta['unit'] = ''
            temp = avail * table
            temp.values[~pd.isnull(temp.values)] = 1
            temp.values[pd.isnull(temp.values)] = 0
            
            avail = avail + temp
        avail = avail / len(self)
        avail = util.cleanDataTable(avail)
        if regionList is not None:
            regionList = avail.index.intersection(regionList)
            avail = avail.loc[regionList,:]
        if years is not None:
            years = avail.columns.intersection(years)
            avail = avail.loc[:,years]
        
        plt.pcolor(avail)
#        plt.clim([0,1])
        plt.colorbar()
        plt.yticks([x +.5 for x in range(len(avail.index))], avail.index)
        plt.xticks([x +.5 for x in range(len(avail.columns))], avail.columns, rotation=45)

class Visualization():
    """ 
    This class addes handy built-in visualizations to datatables
    """
    
    def __init__(self, df):
        self.df = df
    
    def availability(self, regions = None):
        
        if regions is not None:
            available_regions = self.df.index.intersection(regions)
            
            data = self.df.loc[available_regions, :]
        else:
            data = self.df
            
            availableRegions = data.index[~data.isnull().all(axis=1)]
        #print(availableRegions)
        plt.pcolormesh(data, cmap ='RdYlGn_r')
        self._formatTimeCol()
        self._formatSpaceCol(data.index)
        return available_regions
        
    def _formatTimeCol(self):
        years = self.df.columns.values
        
        dt = int(len(years) / 10)+1
            
        xTickts = np.array(range(0, len(years), dt))
        plt.xticks(xTickts+.5, years[xTickts], rotation=45)
        print(xTickts)
        
    def _formatSpaceCol(self, regions = None):
        if regions is None:
            locations = self.df.index.values
        else:
            locations = np.asarray(list(regions))
        
        #dt = int(len(locations) / 10)+1
        dt = 1    
        yTickts = np.array(range(0, len(locations), dt))
        plt.yticks(yTickts+.5, locations[yTickts])

    def plot(self, **kwargs):
        
        if 'ax' not in kwargs.keys():
            if 'ID' in self.df.meta.keys():
                fig = plt.figure(self.df.meta['ID'])
            else:
                fig = plt.figure('unkown')
            ax = fig.add_subplot(111)
            kwargs['ax'] = ax
        self.df.T.plot(**kwargs)
        #print(kwargs['ax'])
        #super(Datatable, self.T).plot(ax=ax)
        kwargs['ax'].set_title(self.df.meta['entity'])
        kwargs['ax'].set_ylabel(self.df.meta['unit'])

    def html_line(self, fileName=None, paletteName= "Category20",returnHandle = False):
        from bokeh.io import show
        from bokeh.plotting import figure
        from bokeh.resources import CDN
        from bokeh.models import ColumnDataSource
        from bokeh.embed import file_html
        from bokeh.embed import components
        from bokeh.palettes import all_palettes
        from bokeh.models import Legend
        tools_to_show = 'box_zoom,save,hover,reset'
        plot = figure(plot_height =600, plot_width = 900,
           toolbar_location='above', tools_to_show=tools_to_show,

        # "easy" tooltips in Bokeh 0.13.0 or newer
        tooltips=[("Name","$name"), ("Aux", "@$name")])
        #plot = figure()

        #source = ColumnDataSource(self)
        palette = all_palettes[paletteName][20]
        
        df = pd.DataFrame([],columns = ['year'])
        df['year'] = self.df.columns
        for spatID in self.df.index:
            df.loc[:,spatID] = self.df.loc[spatID].values
            df.loc[:,spatID + '_y'] = self.df.loc[spatID].values
        
        source = ColumnDataSource(df)
        legend_it = list()
        import datatoolbox as dt
        for spatID,color in zip(self.df.index, palette):
#            coName = mapp.countries.codes.name.loc[spatID]
            coName = dt.mapp.nameOfCountry(spatID)
            #plot.line(x=self.columns, y=self.loc[spatID], source=source, name=spatID)
            c = plot.line('year', spatID + '_y', source=source, name=spatID, line_width=2, line_color = color)
            legend_it.append((coName, [c]))
        plot.legend.click_policy='hide'
        legend = Legend(items=legend_it, location=(0, 0))
        legend.click_policy='hide'
        plot.add_layout(legend, 'right') 
        html = file_html(plot, CDN, "my plot")
        
        if returnHandle: 
            return plot
        
        if fileName is None:
            show(plot)
        else:
            with open(fileName, 'w') as f:
                f.write(html)


    def to_map(self, coList=None, year=None):
        #%%
        import matplotlib.pyplot as plt
        import cartopy.io.shapereader as shpreader
        import cartopy.crs as ccrs
        import matplotlib
        
        df = self.df
        if year is None:
            year = self.df.columns[-1]
        if coList is not None:
            
            df = df.loc[coList,year]
        cmap = matplotlib.cm.get_cmap('RdYlGn')

#        rgba = cmap(0.5)
        norm = matplotlib.colors.Normalize(vmin=df.loc[:,year].min(), vmax=df.loc[:,year].max())
        if 'ID' in list(df.meta.keys()):
            fig = plt.figure(figsize=[8,5], num = self.df.ID)
        else:
            fig = plt.figure(figsize=[8,5])
        ax = plt.axes(projection=ccrs.PlateCarree())
#        ax.add_feature(cartopy.feature.OCEAN)
        
        shpfilename = shpreader.natural_earth(resolution='110m',
                                              category='cultural',
                                              name='admin_0_countries')
        reader = shpreader.Reader(shpfilename)
        countries = reader.records()
        
        for country in countries:
            if country.attributes['ISO_A3_EH'] in df.index:
                ax.add_geometries(country.geometry, ccrs.PlateCarree(),
                                  color = cmap(norm(df.loc[country.attributes['ISO_A3_EH'],year])),
                                  label=country.attributes['ISO_A3_EH'],
                                  edgecolor='white'
                                  )
#            else:
#                ax.add_geometries(country.geometry, ccrs.PlateCarree(),
#                                  color = '#405484',
#                                  label=country.attributes['ISO_A3_EH'])
#        plt.title('Countries that accounted for 95% of coal emissions in 2016')
        
        ax2  = fig.add_axes([0.10,0.05,0.85,0.05])
#        norm = matplotlib.colors.Normalize(vmin=0,vmax=2)
        cb1  = matplotlib.colorbar.ColorbarBase(ax2,cmap=cmap,norm=norm,orientation='horizontal')
        cb1.set_label(self.df.meta['unit'])
        plt.title(self.df.meta['entity'])
        plt.show()
#        plt.colorbar()
#%%
    
    def html_scatter(self, fileName=None, paletteName= "Category20", returnHandle = False):
        from bokeh.io import show
        from bokeh.plotting import figure
        from bokeh.resources import CDN
        from bokeh.models import ColumnDataSource
        from bokeh.embed import file_html
        from bokeh.embed import components
        from bokeh.palettes import all_palettes
        from bokeh.models import Legend
        tools_to_show = 'box_zoom,save,hover,reset'
        plot = figure(plot_height =600, plot_width = 900,
           toolbar_location='above', tools=tools_to_show,
    
        # "easy" tooltips in Bokeh 0.13.0 or newer
        tooltips=[("Name","$name"), ("Aux", "@$name")])
        #plot = figure()
    
        #source = ColumnDataSource(self)
        palette = all_palettes[paletteName][20]
        
        df = pd.DataFrame([],columns = ['year'])
        df['year'] = self.df.columns
        
        for spatID in self.df.index:
            df.loc[:,spatID] = self.df.loc[spatID].values
            df.loc[:,spatID + '_y'] = self.df.loc[spatID].values
        
        source = ColumnDataSource(df)
        legend_it = list()
        import datatoolbox as  dt
        for spatID, color in zip(self.df.index, palette):
            coName = dt.mapp.countries.codes.name.loc[spatID]
            #plot.line(x=self.columns, y=self.loc[spatID], source=source, name=spatID)
            c = plot.circle('year', spatID + '_y', source=source, name=spatID, color = color)
            legend_it.append((coName, [c]))

        legend = Legend(items=legend_it, location=(0, 0))
        legend.click_policy='hide'
        plot.add_layout(legend, 'right')
            #p.circle(x, y, size=10, color='red', legend='circle')
        plot.legend.click_policy='hide'
        html = file_html(plot, CDN, "my plot")
        
        if returnHandle: 
            return plot
        if fileName is None:
            show(plot)
        else:
            with open(fileName, 'w') as f:
                f.write(html)

def _try_number_format(x):  
    try:      
        return int(x)     
    except:
        try:
            return float(x)
        except:
            return x
            
def read_csv(fileName, native_regions=False):
    """
    Load DataTable from csv file.

    Parameters
    ----------
    fileName : str
        Path of file.
    native_regions : bool, optional
            Load native region defintions if available. The default is False.


    Returns
    -------
    DataTable
        DataTable with data and meta.

    """
    
    fid = open(fileName,'r', encoding='utf-8')
    
    assert (fid.readline()) == config.META_DECLARATION
    #print(nMetaData)
    
    meta = dict()
    while True:
        
        line = fid.readline()
        if line == config.DATA_DECLARATION:
            break
        try:
            key, val = line.replace('\n','').split(',', maxsplit=1)
        except:
            continue
        meta[key] = _try_number_format(val.strip())
        if "unit" not in meta.keys():
            meta["unit"] = ""
    
    if 'timeformat' in meta.keys():
         meta['_timeformat'] = meta['timeformat']
         del meta['timeformat']
    # if meta['_timeformat'] == 'Y':
    #     meta['_timeformat'] = '%Y'
    # print(meta) 
    
    df = pd.read_csv(fid)
    if 'standard_region' not in df.columns:
        #backward compatibility
        df = df.set_index(df.columns[0])
    else:
        # new datatable
        if native_regions:
            
            df = df.set_index(['region', 'standard_region'])
            # if :
                # df = df.drop('standard_region',axis=1)
        else:
            df = df.set_index(['standard_region', 'region'])
            df = df.reset_index('region', drop = True)
            df = df[~df.index.isnull()]
    fid.close()
    df = Datatable(df, meta=meta)
    
    if ('_timeformat' in meta.keys()) and (meta['_timeformat'] != '%Y'
                                           and meta['_timeformat'] != 'Y'):
        df.columns_to_datetime()
    else:
        df.columns = df.columns.astype(int)
    
    
    dupl_mask = df.index.duplicated()
    #removing duplicated entries that might results in different native and standart region definitions
    if  sum(dupl_mask) > 0:
        if config.DEBUG:
            print(f'Removing duplicated {sum(dupl_mask)} entry(ies)')
        df = df.iloc[~dupl_mask,:]
    
    return df#.drop_duplicates()

def read_excel(fileName, 
               sheetNames = None,
               use_sheet_name_as_keys=False, 
               force_tableSet = False):
 
    
    if sheetNames is None:
        xlFile = pd.ExcelFile(fileName)
        sheetNames = xlFile.sheet_names
        xlFile.close()

    if len(sheetNames) > 1 or force_tableSet:
        out = TableSet()
        for sheet in sheetNames:
            fileContent = pd.read_excel(fileName, sheet_name=sheet, header=None)
            metaDict = dict()
            try:
                for idx in fileContent.index:
                    key, value = fileContent.loc[idx, [0,1]]
                    if key == '###DATA###':
                        break
                    
                    metaDict[key] = value
                columnIdx = idx +1
                dataTable = Datatable(data    = fileContent.loc[columnIdx+1:, 1:].astype(float).values, 
                                      index   = fileContent.loc[columnIdx+1:, 0], 
                                      columns = [int(x) for x in fileContent.loc[columnIdx, 1:]], 
                                      meta    = metaDict)
                
                if use_sheet_name_as_keys:
                    out[sheet] = dataTable
                else:
                    dataTable.generateTableID()
                    out.add(dataTable)
                
            except Exception:

                
                if config.DEBUG:
                    print(traceback.format_exc())
                print('Failed to read the sheet: {}'.format(sheet))                    
        
    else:
        sheet = sheetNames[0]
        fileContent = pd.read_excel(fileName, sheet_name=sheet, header=None)
        metaDict = dict()
        if True:
            for idx in fileContent.index:
                key, value = fileContent.loc[idx, [0,1]]
                if key == '###DATA###':
                    break
                
                metaDict[key] = value
            columnIdx = idx +1
            dataTable = Datatable(data    = fileContent.loc[columnIdx+1:, 1:].astype(float).values, 
                                  index   = fileContent.loc[columnIdx+1:, 0], 
                                  columns = fileContent.loc[columnIdx, 1:], 
                                  meta    = metaDict)
            if ('_timeformat' in dataTable.meta.keys()) and (dataTable.meta['_timeformat'] != '%Y'):
                dataTable.columns_to_datetime()
        
            try:
                dataTable.generateTableID()
            except:
                print('Warning: Meta data incomplete, table ID not generated')
            out = dataTable
#        except:
#                print('Failed to read the sheet: {}'.format(sheet))
        
    return out

def read_compact_excel(file_name, sheet_name=None):
    #%%
    if sheet_name is None:
        xlFile = pd.ExcelFile(file_name)
        sheet_names = xlFile.sheet_names
        xlFile.close()

    
    out = TableSet()
    for sheet in sheet_names:
        fileContent = pd.read_excel(file_name, sheet_name=sheet, header=None)
        metaDict = dict()
        
        if fileContent.loc[0, 0] != '###META###':
            raise(BaseException('Undefined format'))
            
        # try:
        if True:
            for idx in fileContent.index[1:]:
                key, value = fileContent.loc[idx, [0,1]]
                if key == '###DATA###':
                    break
                
                metaDict[key] = value
            columnIdx = idx +1
            

            lDf = fileContent.loc[columnIdx+1:, :]
            lDf.columns = fileContent.loc[columnIdx, :]
            # lDf.index = fileContent.loc[columnIdx+1:, 0]
            
            if 'ID' in metaDict['meta_columns']:
                lDf.loc[:,'ID'] = fileContent.loc[columnIdx:, 0]
            # import json
            # x = '[ "A","B","C" , " D"]'
            # json.loads(metaDict['meta_columns'])
            
            try:
                meta_columns = ast.literal_eval(metaDict['meta_columns'])
            except:
                meta_columns = metaDict['meta_columns']
            if isinstance(meta_columns, str):
                meta_columns= meta_columns.split(', ')
            # if 'region' in meta_columns:
            meta_columns.remove('region')
                
            for idx, line in lDf.iterrows():
                
                #print(idx)
                    
                meta = { x: metaDict[x] for x in metaDict.keys() if x not in  ['meta_columns']}
                
                for meta_col in meta_columns:
                    if meta_col in line.keys():
                        meta[meta_col] = line[meta_col]
                region = line.loc['region']
                
                if 'ID' in line.index:
                    ID = line.loc['ID']
                else:
                    try:
                        meta = core._update_meta(meta)
                        ID = core._createDatabaseID(meta)
                    except:
                        
                        ID = config.ID_SEPARATOR.join([meta[key] for key in [x for x in config.ID_FIELDS if x in meta.keys()]])
                        print(f'ID could not properly created, fallback to partial ID: {ID}')
                if config.DEBUG:
                    print(f'meta columns: {meta_columns}')
                    print(f'numerical columns: {list(lDf.columns.difference(meta_columns + ["region"]))}')
                numerical_columns = list(lDf.columns.difference(meta_columns + ['region']).astype(int))
                
                if ID not in out.keys():
                    # table  = Datatable()
                    table = Datatable(
                                columns= numerical_columns,
                                index= [region],                
                                meta = meta)
                    
                    table.loc[region, numerical_columns] = line.loc[numerical_columns].astype(float)
                    out[ID] = table
                    
                else:
                    line.loc[numerical_columns].astype(float)
                    factor = core.conversionFactor(meta['unit'], out[ID].meta['unit'])
                    out[ID].loc[region, numerical_columns] = line.loc[numerical_columns].astype(float) *factor
                    #
         
    return out
            
        # except:
        #         print('Failed to read the sheet: {}'.format(sheet))


#%%
class MetaData(dict):
    
    def __init__(self):
        super(MetaData, self).__init__()
        self.update({x : '' for x in config.REQUIRED_META_FIELDS})
    
    
    def __setitem__(self, key, value):
        super(MetaData, self).__setitem__(key, value)
        super(MetaData, self).__setitem__('variable', '|'.join([self[key] for key in ['entity', 'category'] if key in self.keys()]))
        super(MetaData, self).__setitem__('pathway', '|'.join([self[key] for key in ['scenario', 'model'] if key in self.keys()]))
        super(MetaData, self).__setitem__('source', '_'.join([self[key] for key in ['institution', 'year'] if key in self.keys()]))
        
#%% Test     
def _add_required_meta(data, meta, stacked_dims):
    
    for key in stacked_dims.keys():
        
        dims_to_add = list()
        for dim in stacked_dims[key]:
            if dim in data.index.names:
                continue
            else:
                if dim in meta.columns:
                    dims_to_add.append(dim)
        data = data.join(meta.loc[:, dims_to_add])
        idx_names = set(data.index.names).union(dims_to_add)
        print(dims_to_add)
        print(idx_names)
        data = data.reset_index().set_index(list(idx_names))
        return data
    
def _add_meta(data, meta, stacked_dims):
    
    data = data.join(meta)
    # data['meta'] = range(len(data.index))
    stacked_dims.update({'meta' : list(meta.columns)}) 
    idx_names = set(data.index.names).union(list(meta.columns))
    data = data.reset_index().set_index(list(idx_names))
    return data, stacked_dims            

def _pack_dimensions(index, **stacked_dims):
    packed_labels = {}
    packed_values = {}
    drop_levels = []
    
    for dim, levels in stacked_dims.items():
        labels = pd.MultiIndex.from_arrays([index.get_level_values(l) for l in levels])
        packed_labels[dim] = labels_u = labels.unique()
        packed_values[dim] = pd.Index(labels_u.get_indexer(labels), name=dim)
        drop_levels.extend(levels)

    return (
        pd.MultiIndex.from_arrays(
            [index.get_level_values(l) for l in index.names.difference(drop_levels)] +
            list(packed_values.values())
        ),
        packed_labels
    )

# def _pack_dimensions(index, 
#                      meta = None, 
#                      **stacked_dims):
#     packed_labels = {}
#     packed_values = {}
#     drop_levels = []
    
#     for dim, levels in stacked_dims.items():
#         labels = pd.MultiIndex.from_arrays([index.get_level_values(l) for l in levels])
#         packed_labels[dim] = labels_u = labels.unique()
#         packed_values[dim] = pd.Index(labels_u.get_indexer(labels), name=dim)
#         drop_levels.extend(levels)

#     #meta 
#     if meta is not None:
#         for st_dim in stacked_dims.keys():
#             if st_dim in meta.columns:
#                 meta = meta.copy().drop(st_dim,axis= 1)
#         labels = pd.MultiIndex.from_arrays([index.get_level_values(l) for l in meta.index.names])
#         labels_u = labels.unique()
#         packed_values['meta'] = pd.Index(labels_u.get_indexer(labels), name='meta')
#         packed_labels['meta'] = pd.MultiIndex.from_arrays(meta.loc[labels_u,:].values.T, names = meta.columns)
    
    
#     return (
#         pd.MultiIndex.from_arrays(
#             [index.get_level_values(l) for l in index.names.difference(drop_levels)] +
#             list(packed_values.values())
#         ),
#         packed_labels
#     )

import xarray as xr
class DataSet(xr.Dataset):
    """
    Very simple class to allow initialization of xarray datasets from pyam, wide pandas dataframes 
    and datatoolbox queries.
    """
    __slots__ = (
        '_attrs',
         '_cache',
         '_coord_names',
         '_dims',
         '_encoding',
         '_close',
         '_indexes',
         '_variables')
    
    @classmethod
    def from_wide_dataframe(cls, 
                            data,
                            meta = None,
                            stacked_dims = {'pathway' : ("model", "scenario")}):
        to_merge = list()
        for (variable, unit), df_ in data.groupby(['variable', 'unit']):
            #print(df_)
            
            index = df_.index
            df_ = df_.reset_index(['variable', 'unit'], drop = True)
            index, dims = _pack_dimensions(df_.index, **stacked_dims)
            da = (xr.DataArray(df_.set_axis(index)
                  .rename_axis(columns="year"))
                  .unstack("dim_0")
                  .pint.quantify(unit)
                  .assign_coords(dims)
                  )
            to_merge.append(xr.Dataset({variable : da}))
        self = xr.merge(to_merge)
        return self
      
    @classmethod
    def from_pyam(cls, 
                  data,
                  meta_dims =None,
                  stacked_dims = {'pathway' : ("model", "scenario")}):
        
        if meta_dims is not None:
            meta = data.meta.loc[:,meta_dims]
            data, stacked_dims = _add_meta(data, meta, stacked_dims)
        else:
            meta = None
        data = data.timeseries()
        #data = _add_required_meta(data, meta, stacked_dims)
        
        return cls.from_wide_dataframe(data, meta, stacked_dims)
        
    @classmethod
    def from_query(cls,
                   query,
                   stacked_dims = {'pathway' : ("model", "scenario")}):
        dimensions = ['model', 'scenario', 'region', 'variable', 'source', 'unit']
        data = query.as_wide_dataframe(meta_list = dimensions).set_index(dimensions)
        
        return cls.from_wide_dataframe(data, meta = None, stacked_dims = stacked_dims)
    
    
if __name__ == '__main__':
    meta = MetaData()
    meta['entity'] = 'Emissions|CO2'
    meta['institution'] = 'WDI'
    meta['year']  = '2020'
    #print(meta)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu May 14 19:43:58 2020

@author: ageiges
"""

import datatoolbox as dt
import numpy as np
import pandas as pd


metaDict = {'source'   : 'TEST',
            'entity'   : 'values',
            'category' : 'cat1',
            'scenario' : '#1',
            'unit'     : 'm'}
metaDict2 = {'source' : 'TEST2',
            'entity'  : 'area',
            'category': 'cat2',
            'scenario': '#2',
            'unit'    : 'km'}    
    
data = np.asarray([[1,2.2],
                   [2.3, np.nan],
                   [1.3, np.nan],
                   [np.nan, 3.4, ]])


data2 = np.asarray([[1,2.2],
               [2.3, np.nan],
               [1.1, np.nan ],
               [np.nan, 3.3,]])


df = dt.Datatable(data, meta=metaDict, columns = [2010, 2012], index = ['ARG', 'DEU', 'FRA', 'GBR'])
df2 = dt.Datatable(data2, meta=metaDict2, columns = [2009, 2012], index = ['ARG', 'DEU', 'FRA', 'GBR'])
    
    

def test_basics():
    
    
    tableSet = dt.TableSet()
    
    assert hasattr(tableSet,'inventory')
    
    
def test_add_remove():
    tableSet = dt.TableSet()
    tableSet.add(df)
    
    obs = list(tableSet.keys())
    exp = ['values|cat1__#1__TEST']
    assert obs[0] == exp[0]
    
    tableSet.add(df2)
    assert len(tableSet.inventory.index) == 2
    tableSet.remove(df.ID)
    
    obs = list(tableSet.keys())
    exp = ['area|cat2__#2__TEST2']
    assert obs == exp
    
    assert len(tableSet.inventory.index) == 1

def test_add_list():
    tableSet = dt.TableSet()
    tableSet.add([df, df2])
    obs = list(tableSet.keys())
    exp = ['values|cat1__#1__TEST', 'area|cat2__#2__TEST2']
    assert obs == exp
    
def test_to_long_table():
    
    tableSet = dt.TableSet()
    tableSet.add([df, df2])
    
    longTable = tableSet.to_LongTable()

    exp = pd.DataFrame.from_dict(
            {'variable': {0: 'values|cat1',
              1: 'values|cat1',
              2: 'values|cat1',
              3: 'values|cat1',
              4: 'area|cat2',
              5: 'area|cat2',
              6: 'area|cat2',
              7: 'area|cat2'},
            'region': {0: 'ARG',
              1: 'DEU',
              2: 'FRA',
              3: 'GBR',
              4: 'ARG',
              5: 'DEU',
              6: 'FRA',
              7: 'GBR'},
            'scenario': {0: '#1',
              1: '#1',
              2: '#1',
              3: '#1',
              4: '#2',
              5: '#2',
              6: '#2',
              7: '#2'},
            'model': {0: '', 1: '', 2: '', 3: '', 4: '', 5: '', 6: '', 7: ''},
             'unit': {0: 'm', 1: 'm', 2: 'm', 3: 'm', 4: 'km', 5: 'km', 6: 'km', 7: 'km'},
             2009: {0: np.nan, 1: np.nan, 2: np.nan, 3: np.nan, 4: 1.0, 5: 2.3, 6: 1.1, 7: np.nan},
             2010: {0: 1.0, 1: 2.3, 2: 1.3, 3: np.nan, 4: np.nan, 5: np.nan, 6: np.nan, 7: np.nan},
             2012: {0: 2.2, 1: np.nan, 2: np.nan, 3: 3.4, 4: 2.2, 5: np.nan, 6: np.nan, 7: 3.3},
             }
            )
    # exp = exp.loc[:,longTable.columns]
    
    assert (longTable.values[~pd.isna(longTable)] == exp.values[~pd.isna(exp)]).all()

def test_supplement_material():
    tableSet = dt.TableSet()
    tableSet.add([df, df2])

    assert  tableSet.scenarios() == ['#1', '#2']
    assert  tableSet.entities()  == ['values', 'area']
    assert  tableSet.sources()   == ['TEST', 'TEST2']
    
    
if __name__ == '__main__':    
    test_basics()
    test_add_remove()
    test_add_list()
    test_to_long_table()
    test_supplement_material()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: Andreas Geiges

TUTORIAL:       
      
"""

import datatoolbox as dt
import pandas as pd
import matplotlib.pyplot as plt
#%% simple conversion factory

# conversion form kt oil equivalent to Exajoule
dt.conversionFactor('ktoe', 'EJ')

# conversion of compont units 
dt.conversionFactor('ktoe / yr', 'EJ / d')


# conversion of tables
finalEnergyCommercial = dt.getTable('Final_Energy|Comercial_and_public|Total__Historic__IEA_WEB_2019')
print("Final energy Commercial buildings")
print("Germany: {:2.2f} {}".format(finalEnergyCommercial.loc['DEU',2015],finalEnergyCommercial.meta['unit']))

finalEnergyCommercial = finalEnergyCommercial.convert('TWh / d')
print("Germany: {:2.2f} {}".format(finalEnergyCommercial.loc['DEU',2015],finalEnergyCommercial.meta['unit']))


finalEnergyResidential = dt.getTable('Final_Energy|Residential|Total__Historic__IEA_WEB_2019')
print("Final energy Residential buildings")
print("Germany: {:2.2f} {}".format(finalEnergyResidential.loc['DEU',2015],finalEnergyResidential.meta['unit']))


#%% computing with tables
# unit of first table is used to convert tables for calculations 
# missmatch of units does not lead to errors

totalFinalEnergy = finalEnergyCommercial + finalEnergyResidential
print("Final energy: Buildings total")
print("Germany: {:2.2f} {}".format(totalFinalEnergy.loc['DEU',2015],totalFinalEnergy.meta['unit']))

#%% example of impossible conversions


#%% Green house warming potentail conversion
table_N2O_Agri = dt.getTable('Emissions|N2O|Agriculture__Historic__FAO_2019')

#table_N2O_Agri = table_N2O_Agri.loc[countryList,years]
print(table_N2O_Agri)


table_CH4_Agri = dt.getTable('Emissions|CH4|Agriculture__Historic__FAO_2019')#.loc[countryList,years]
print(table_CH4_Agri)

# not working
table_CH4_Agri + table_N2O_Agri

# conversion with GWP context to CO2eq
table_CH4_Agri = table_CH4_Agri.convert('Mt CO2eq', context='GWPAR4')
table_N2O_Agri = table_N2O_Agri.convert('Mt CO2eq', context='GWPAR4')
table_N2O_CH4_Agri = table_CH4_Agri + table_N2O_Agri
table_N2O_CH4_Agri






#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Mar 19 09:58:27 2019

@author: Andreas Geiges
"""
import re
import os
import pandas as pd
import numpy as np
from tqdm import tqdm
import deprecated as dp

from . import config

from . import core
from . import util
from . import mapping as mapp
from .tools.pandas import yearsColumnsOnly

 
from . import greenhouse_gas_database as gh
GHG_data = gh.GreenhouseGasTable()
from .data_structures import Datatable,TableSet
from openpyxl import load_workbook

REQUIRED_SETUP_FIELDS = ['filePath',
                         'fileName',
                         'sheetName',
                         'timeIdxList',
                         'spaceIdxList']

SP_ARG = 3
TI_ARG = 2
DT_ARG = 4

from datatoolbox.tools.excel import ExcelReader

#%% Functions

from .tools import excel as _xls

excel_to_pandas_idx = _xls.excel_to_pandas_idx
getAllSheetNames   = _xls.getAllSheetNames


def read_MAGICC6_MATLAB_bulkout(pathName):
    
    temp_offset = 0.61
    df = pd.read_table(pathName, skiprows=23, header=0, delim_whitespace=True, index_col=0)
    df = df + temp_offset
    df.index = df.index.astype(int)
    return df
    #df.values[df.values<0] = np.nan
#    
#    meta = dict()
#    meta['entity'] = 'global_temp'
#    meta['unit']   = '°C'
#    return Datatable(df, meta=meta)

def read_MAGICC6_BINOUT(filePath):
    import pymagicc as pym
    reader = pym.io._BinaryOutReader(filePath)

    metaData, df = reader.read()
    
    data = df.pivot(index='region',columns='time', values='value')
    
    metaDict = dict()
    metaDict['entity'] = df.variable[0]
    metaDict['source'] = 'MAGICC6_calculation'
    metaDict['unit']   = None

    return Datatable(data, meta=metaDict)

def read_PRIMAP_csv(fileName):
    
    metaMapping = {
            'entity': 'SHEET_ENTITY',
            'unit':'SHEET_UNIT',
            'category':'SHEET_NAME_CATEGORY',
            'scenario': 'SHEET_SCENARIO',
            'model' : 'SHEET_SOURCE'
            }
    
    allDf = pd.read_csv(fileName, usecols=[0,1], index_col=0)
    #print(os.path.basename(fileName))

    firstDataRow = allDf.loc['SHEET_FIRSTDATAROW','Unnamed: 1']
    
    #bugfix of old wrong formated PRIMAP files
    try:
        int(firstDataRow)
    except:
        firstDataRow = pd.np.where(allDf.index == "Countries\Years")[0][0]+3
   
    firstMetaRow = pd.np.where(allDf.index == "&SHEET_SPECIFICATIONS")[0][0]+1
    
    metaPrimap = dict()
    for row in range(firstMetaRow,firstDataRow):
        key = allDf.index[row]
        value = allDf.iloc[row,0]
        if key =='/':
            break
        metaPrimap[key] = value
    
    data = pd.read_csv(fileName, header=firstDataRow-2, index_col=0)
    
    meta = dict()
    for metaKey in metaMapping:
        
        if isinstance(metaMapping[metaKey], list):
            value = '_'.join(metaPrimap[x] for x in metaMapping[metaKey])
        else:
            value = metaPrimap[metaMapping[metaKey]]
        meta[metaKey] = value
    
    
    
    table = Datatable(data, meta=meta)
    table = table.loc[:, yearsColumnsOnly(table)]
    table.columns = table.columns.astype(int)
    return table#, data

def read_PRIMAP_Excel(fileName, 
                      sheet_names= None,
                      sheet_name = None,
                      xlsFile=None):
    
    #single sheet reading
    if (sheet_names is None) and (sheet_name is not None):
        return _read_PRIMAP_Excel_single(fileName, sheet_name, xlsFile)
    
    #single sheet reading
    if isinstance(sheet_names, str) :
        return _read_PRIMAP_Excel_single(fileName, sheet_names, xlsFile)
    
    if sheet_names is None:
        sheet_names = getAllSheetNames(fileName)
        
    out = TableSet()
    if xlsFile is None:
        xlsFile = pd.ExcelFile(fileName)
    for sheet_name in tqdm(sheet_names):
        table = _read_PRIMAP_Excel_single(fileName, sheet_name, xlsFile=xlsFile)
        out[sheet_name] = table
        
    return out
        
def _read_PRIMAP_Excel_single(fileName, sheet_name= 0, xlsFile=None):
    if xlsFile is None: 
        xlsFile = pd.ExcelFile(fileName)
    allDf = pd.read_excel(xlsFile, sheet_name = sheet_name, usecols=[0,1], index_col=0)
    #print(os.path.basename(fileName))

    firstDataRow = allDf.loc['SHEET_FIRSTDATAROW','Unnamed: 1']
    
    #bugfix of old wrong formated PRIMAP files
    try:
        int(firstDataRow)
    except:
        firstDataRow = pd.np.where(allDf.index == "Countries\Years")[0][0]+3
        
    #print(firstDataRow)
    setup = dict()
    setup['filePath']  = os.path.dirname(fileName) +'/'
    setup['fileName']  = os.path.basename(fileName)
    setup['sheetName'] = sheet_name
    setup['timeIdxList']  = ('B' + str(firstDataRow-1), 'XX'+ str(firstDataRow-1))
    setup['spaceIdxList'] = ('A'+ str(firstDataRow), 'A1000')
    #print(setup)
    ex = ExcelReader(setup, xlsFile=xlsFile)
    data = ex.gatherData().astype(float)
    #return data
    meta = {'source': '',
            'entity': allDf.loc['SHEET_ENTITY','Unnamed: 1'],
            'unit': allDf.loc['SHEET_UNIT','Unnamed: 1'],
            'category': allDf.loc['SHEET_NAME_CATEGORY','Unnamed: 1'],
            'scenario': allDf.loc['SHEET_SCENARIO','Unnamed: 1'] + '|' + allDf.loc['SHEET_SOURCE','Unnamed: 1']}
    REG_ton = re.compile('^[GM]t')
    xx = REG_ton.search(meta['unit'])
    
    if xx:
        meta['unit'] = meta['unit'].replace(xx.group(0), xx.group(0) + ' ')
    
    table = Datatable(data, meta=meta)
    try:
        table = table.loc[:,yearsColumnsOnly(table)]
        table.columns = table.columns.astype(int)
#        table = table.loc[:,yearsColumnsOnly(table)]
    except:
        print('warning: Columns could not be converted to int')
    return table

def read_MAGICC6_ScenFile(fileName, **kwargs):
    VALID_MASS_UNITS= {
            'Pt': 1e18,
            'Gt': 1e15,
            'Mt': 1e12,
            'kt': 1e9,
            't' : 1e6,
            'Pg': 1e15,
            'Tg': 1e12,
            'Gg': 1e9,
            'Mg': 1e6,
            'kg': 1e3,
            'g' : 1}
    fid = open(fileName,'r')
    nDataRows    = int(fid.readline().replace('/n',''))
    
    while True:
    #for i, line in enumerate(fid.readlines()):
        line = fid.readline()
        if line[:11] == '{0: >11}'.format('YEARS'):
            break
    #get first header line
    entities  = line.split()[1:]
    
    
    #reading units
    unitLine = fid.readline().split()[1:]
    #print(unitLine)
    
    # find correct component
    components = [GHG_data.findEntryIdx(entity) for entity in entities]

    units = [unit for unit in unitLine if unit[:2] in VALID_MASS_UNITS]        
    
    
    replacementDict = {'MtN2O-N' : 'Mt N'}

    units = [replacementDict.get(unit, unit) for unit in units]

    columns = [(x,y,z) for x,y,z in zip(entities, components, units)]
    entityFrame =  pd.DataFrame(columns=entities)
    entityFrame.columns = pd.MultiIndex.from_tuples(columns)
    
    entityFrame.columns.names=['NAME', 'COMP','UNIT']

    for i, line in enumerate(fid.readlines()):
        
        if i == nDataRows:
            break
        data = line.split()
        entityFrame.loc[int(data[0])] = np.asarray(data[1:])

    # TODO: CHange to a results list of datatables
    return entityFrame

def insertDataIntoExcelFile(fileName, 
                            overwrite = False, 
                            setupSheet = 'INPUT',
                            interpolate = False):
    ins = _xls.ExcelWriter(fileName=fileName, overwrite = overwrite, setupSheet = setupSheet, interpolate=interpolate)
    ins.insert_data()
    ins.close()
    return ins
    

    
getAllSheetNames = _xls.getAllSheetNames


     
#%%
if False:
    #%%
    ws_readValues = load_workbook(self.setup['fileName'], data_only=True)['Sheet6']
    timeIdxString  = 'B2:K2'
    spaceIdxString = 'A4:A8'
    dataIdxString  = 'population|all|PROJECTION_MED|UN_WPP2017'
    
    ws_readValues = load_workbook(self.setup['fileName'], data_only=True)['Sheet7']
    timeIdxString  = 'B1:O1'
    spaceIdxString = 'DEU'
    dataIdxString  = 'A2:A6'
    
    ws_readValues = load_workbook(self.setup['fileName'], data_only=True)['Sheet8']
    timeIdxString  = '2020'
    spaceIdxString = 'B1:O1'
    dataIdxString  = 'A2:A6'

#Sheet2	B2:E2	A3:A4	
    ws_readValues = load_workbook(self.setup['fileName'], data_only=True)['Sheet2']
    timeIdxString  = 'B2:E2'
    spaceIdxString = 'A3:A4'
    dataIdxString  = 'area_agriculture|agriculture|historic|WDI2018'
##Sheet2	B2:E2	A3:A4	
#    ws_readValues = load_workbook(self.setup['fileName'], data_only=True)['Sheet4']
#    timeIdxString  = 'A'
#    spaceIdxString = 'B2:D2'
#    dataIdxString  = 'population|all|PROJECTION_LOW|UN_WPP2017'

    ws_readValues = load_workbook(self.setup['fileName'], data_only=True)['Sheet4a']
    timeIdxString  = 'A'
    spaceIdxString = '2'
    dataIdxString  = 'population|all|PROJECTION_LOW|UN_WPP2017'

    
    args =  None, None, None, None, None
    tableIDs = list()
    for args in iterData(dataIdxString, ws_readValues, *args):
        tableIDs.append(args[4])
    tables = dt.getTables(tableIDs)
    wksSheet = ws_readValues
    args =  None, None, None, None, None
    for argsTime in iterTime(timeIdxString, wksSheet, *args):
        if argsTime[TI_ARG] is None:
            continue
        for argsSpace in iterSpace(spaceIdxString, wksSheet, *argsTime):
            if argsSpace[SP_ARG] is None:
                continue
            for argData in iterData(dataIdxString, wksSheet, *argsSpace):
                if argData[DT_ARG] is None:
                    continue
                print(argData)
                try:
                    value = tables[argData[DT_ARG]].loc[argData[SP_ARG],int(argData[TI_ARG])]
                    print('success')
                except:
                    pass
#%%


def PandasExcelWriter(fileName):
    return pd.ExcelWriter(fileName,
                        engine='xlsxwriter',
                        datetime_format='mmm d yyyy hh:mm:ss',
                        date_format='mmmm dd yyyy')


#%%


        
if __name__ == '__main__':
    import datatools as dt
#    setup = dict()
#    setup['filePath']  = '../../projects/NDC-analysis/03_Ready/'
#    setup['fileName']  = 'CAT_CountryAssessment_Argentina_201812.xlsx'
#    setup['sheetName'] = 'OutputWebsite'
#    setup['timeIdxList']  = ('L40', 'AZ40')
#    setup['spaceIdxList'] = ('I42', 'I42')
#    
#    ex = Extractor(setup)
#    
#    print(ex.timeIdxList)
#    timeSeries = ex.gatherData()
#    
#    ex.setup(**{'spaceIdxList': ('I42', 'I43')})
#    dataDf = ex.gatherData()
#
#    ex.setup(**{'timeIdxList': ('L40', 'L40')})
#    spatSeries = ex.gatherData()    
#
#    ex.setup(**{'spaceIdxList': ('I42', 'I42')})
#    value = ex.gatherData() 
#    
#    excel_to_pandas_idx('AB12')    
    
    #%%
    from shutil import copyfile
    copyfile('demo_empty.xlsx', "demo.xlsx")
    ins = ExcelWriter(fileName='demo.xlsx')
    tables = ins.insert_data()
    ins.close()
    #%%    
    reader = ExcelReader_New(fileName='demo_filled.xlsx')
    out = reader.read_data()
#    for setup in ins.getSetups():
#        dataTable = dt.getTable(setup['dataID'])
#        ins._writeData(setup, dataTable)